# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Martin Drawitsch, Philipp Schubert

__all__ = ['PatchCreator', 'SimpleNeuroData2d', 'Segmentation2d', 'Reconstruction2d']

import glob
import logging
import os
import sys
import traceback
from os.path import expanduser
from typing import Tuple, Dict, Optional, Union, Sequence, Any, List, Callable

import h5py
import imageio
import numpy as np
import torch
from torch.utils import data

from elektronn3.data import coord_transforms, transforms
from elektronn3.data.sources import DataSource, HDF5DataSource, slice_3d

logger = logging.getLogger('elektronn3log')


class _DefaultCubeMeta:
    def __getitem__(self, *args, **kwargs): return np.inf


# TODO: Document passing DataSources directly
class PatchCreator(data.Dataset):
    """Dataset iterator class that creates 3D image patches from HDF5 files.

    It implements the PyTorch ``Dataset`` interface and is meant to be used
    with a PyTorch ``DataLoader`` (or the modified
    :py:class:`elektronn3.training.trainer.train_utils.DelayedDataLoader`, if it is
    used with :py:class:`elektronn3.training.trainer.Trainer``).

    The main idea of this class is to automate input and target patch creation
    for training convnets for semantic image segmentation. Patches are sliced
    from random locations in the supplied HDF5 files (``input_h5data``,
    ``target_h5data``).
    Optionally, the source coordinates from which patches
    are sliced are obtained by random warping with affine or perspective
    transformations for efficient augmentation and avoiding border artifacts
    (see ``warp_prob``, ``warp_kwargs``).
    Note that whereas other warping-based image augmentation systems usually
    warp images themselves, elektronn3 performs warping transformations on
    the **coordinates** from which image patches are sliced and obtains voxel
    values by interpolating between actual image voxels at the warped source
    locations (which are not confined to the original image's discrete
    coordinate grid).
    (TODO: A visualization would be very helpful here to make this more clear)
    For more information about this warping mechanism see
    :py:meth:`elektronn3.data.cnndata.warp_slice()`.

    Currently, only 3-dimensional image data sets are supported, but 2D
    support is also planned.

    Args:
        input_sources: Sequence of ``(filename, hdf5_key)`` tuples, where
            each item specifies the filename and
            the HDF5 dataset key under which the input data is stored.
        target_sources: Sequence of ``(filename, hdf5_key)`` tuples, where
            each item specifies the filename and
            the HDF5 dataset key under which the target data is stored.
        patch_shape: Desired spatial shape of the samples that the iterator
            delivers by slicing from the data set files.
            Since this determines the size of input samples that are fed
            into the neural network, this is a very important value to tune.
            Making it too large can result in slow training and excessive
            memory consumption, but if it is too small, it can hinder the
            perceptive ability of the neural network because the samples it
            "sees" get too small to extract meaningful features.
            Adequate values for ``patch_shape`` are highly dependent on the
            data set ("How large are typical ROIs? How large does an image
            patch need to be so you can understand the input?") and also
            depend on the neural network architecture to be used (If the
            effective receptive field of the network is small, larger patch
            sizes won't help much).
        offset: Shape of the offset by which each the targets are cropped
            on each side. This needs to be set if the outputs of the network
            you train with are smaller than its inputs.
            For example, if the spatial shape of your inputs is
            ``patch_shape=(48, 96, 96)`` the spatial shape of your outputs is
            ``out_shape=(32, 56, 56)``, you should set ``offset=(8, 20, 20)``,
            because ``offset = (patch_shape - out_shape) / 2`` should always
            hold true.
        cube_prios: List of per-cube priorities, where a higher priority
            means that it is more likely that a sample comes from this cube.
        aniso_factor: Depth-anisotropy factor of the data set. E.g.
            if your data set has half resolution in the depth dimension,
            set ``aniso_factor=2``. If all dimensions have the same
            resolution, set ``aniso_factor=1``.
        input_discrete_ix: List of input channels that contain discrete values.
            By default (``None``), no channel is seen as discrete (generally
            inputs are real world images).
            This information is used to decide what kind of interpolation should
            be used for reading input data:
            - discrete targets are obtained by nearest-neighbor interpolation
            - non-discrete (continuous) targets are linearly interpolated.
        target_discrete_ix: List of target channels that contain discrete values.
            By default (``None``), every channel is seen as discrete (this is
            generally the case for classification tasks).
            See input_discrete_ix for the effect on target interpolation.
        target_dtype: dtype that target tensors should be cast to.
        train: Determines if samples come from training or validation
            data.
            If ``True``, training data is returned.
            If ``False``, validation data is returned.
        warp_prob: ratio of training samples that should be obtained using
            geometric warping augmentations.
        warp_kwargs: kwargs that are passed through to
            :py:meth:`elektronn3.data.coord_transforms.get_warped_slice()`.
            See the docs of this function for information on kwargs options.
            Can be empty.
        epoch_size: Determines the length (``__len__``) of the ``Dataset``
            iterator. ``epoch_size`` can be set to an arbitrary value and
            doesn't have any effect on the content of produced training
            samples. It is recommended to set it to a suitable value for
            one "training phase", so after each ``epoch_size`` batches,
            validation/logging/plotting are performed by the training loop
            that uses this data set (e.g.
            ``elektronn3.training.trainer.Trainer``).
        transform: Transformation function to be applied to ``(inp, target)``
            samples (for normalization, data augmentation etc.). The signature
            is always ``inp, target = transform(inp, target)``, where ``inp``
            and ``target`` both are ``numpy.ndarray``s.
            In some transforms ``target`` can also be set to ``None``. In this
            case it is ignored and only ``inp`` is processed.
            To combine multiple transforms, use
            :py:class:`elektronn3.data.transforms.Compose`.
            See :py:mod:`elektronn3.data.transforms`. for some implementations.
        in_memory: If ``True``, all data set files are immediately loaded
            into host memory and are permanently kept there as numpy arrays.
            If this is disabled (default), file contents are always read from
            the HDF5 files to produce samples. (Note: This does not mean it's
            slower, because file contents are transparently cached by h5py,
            see http://docs.h5py.org/en/latest/high/file.html#chunk-cache).

    """
    def __init__(
            self,
            input_sources: List[Tuple[str, str]],
            patch_shape: Sequence[int],
            target_sources: Optional[List[Tuple[str, str]]] = None,
            offset: Sequence[int] = (0, 0, 0),
            cube_prios: Optional[Sequence[float]] = None,
            aniso_factor: int = 2,
            target_discrete_ix: Optional[List[int]] = None,
            input_discrete_ix: Optional[List[int]] = None,
            target_dtype: np.dtype = np.int64,
            train: bool = True,
            warp_prob: Union[bool, float] = False,
            warp_kwargs: Optional[Dict[str, Any]] = None,
            epoch_size: int = 100,
            transform: Callable = transforms.Identity(),
            in_memory: bool = False,
            cube_meta=_DefaultCubeMeta(),
    ):
        # Early checks
        if target_sources is not None and len(input_sources) != len(target_sources):
            raise ValueError(
                'If target_sources is not None, input_sources and '
                'target_sources must be lists of same length.'
            )
        if not train:
            if warp_prob > 0:
                logger.warning('Augmentations should not be used on validation data.')

        # batch properties
        self.train = train
        self.warp_prob = warp_prob
        self.warp_kwargs = warp_kwargs if warp_kwargs is not None else {}

        # general properties
        self.input_sources = input_sources
        self.target_sources = target_sources
        self.cube_meta = cube_meta
        self.cube_prios = cube_prios
        self.aniso_factor = aniso_factor
        self.target_discrete_ix = target_discrete_ix
        self.input_discrete_ix = input_discrete_ix
        self.epoch_size = epoch_size
        self._orig_epoch_size = epoch_size  # Store original epoch_size so it can be reset later.
        self.in_memory = in_memory

        self.patch_shape = np.array(patch_shape, dtype=np.int)
        self.ndim = self.patch_shape.ndim
        self.offset = np.array(offset)
        self.target_patch_shape = self.patch_shape - self.offset * 2
        self._target_dtype = target_dtype
        self.transform = transform

        # Setup internal stuff
        self.pid = os.getpid()

        # The following fields will be filled when reading data
        self.n_labelled_pixels = 0
        self.inputs: List[DataSource] = []
        self.targets: List[DataSource] = []

        self.load_data()  # Open dataset files

        self.n_successful_warp = 0
        self.n_failed_warp = 0
        self._failed_warp_warned = False

    def __getitem__(self, index: int) -> Dict[str, Any]:
        # Note that the index is ignored. Samples are always random
        return self._get_random_sample()

    def _get_random_sample(self) -> Dict[str, Any]:
        input_src, target_src, i = self._getcube()  # get cube randomly
        warp_prob = self.warp_prob
        while True:
            try:
                inp, target = self.warp_cut(input_src, target_src, warp_prob, self.warp_kwargs)
                if target is not None:
                    target = target.astype(self._target_dtype)
            except coord_transforms.WarpingOOBError as e:
                # Temporarily set warp_prob to 1 to make sure that the next attempt
                #  will also try to use warping. Otherwise, self.warp_prob would not
                #  reflect the actual probability of a sample being obtained by warping.
                warp_prob = 1 if warp_prob > 0 else 0
                self.n_failed_warp += 1
                if self.n_failed_warp > 20 and self.n_failed_warp > 8 * self.n_successful_warp and not self._failed_warp_warned:
                    fail_ratio = self.n_failed_warp / (self.n_failed_warp + self.n_successful_warp)
                    fail_percentage = int(round(100 * fail_ratio))
                    print(e)
                    logger.warning(
                        f'{fail_percentage}% of warping attempts are failing.\n'
                        'Consider lowering lowering your input patch shapes or warp_kwargs[\'warp_amount\']).'
                    )
                    self._failed_warp_warned = True
                continue
            except coord_transforms.WarpingSanityError:
                logger.exception('Invalid coordinate values encountered while warping. Retrying...')
                continue
            self.n_successful_warp += 1
            try:
                inp, target = self.transform(inp, target)
            except transforms._DropSample:
                # A filter transform has chosen to drop this sample, so skip it
                logger.debug('Sample dropped.')
                continue
            break

        inp = torch.as_tensor(inp)
        cube_meta = torch.as_tensor(self.cube_meta[i])
        fname = os.path.basename(self.inputs[i].fname)
        sample = {
            'inp': inp,
            'cube_meta': cube_meta,  # TODO: Make cube_meta completely optional again
            'fname': fname
        }
        if target is not None:
            sample['target'] = torch.as_tensor(target)

        return sample

    def __len__(self) -> int:
        return self.epoch_size

    # TODO: Write a good __repr__(). The version below is completely outdated.
    # def __repr__(self) -> str:
    #     s = "{0:,d}-target Data Set with {1:,d} input channel(s):\n" + \
    #         "#train cubes: {2:,d} and #valid cubes: {3:,d}, {4:,d} labelled " + \
    #         "pixels."
    #     s = s.format(self.c_target, self.c_input, self._training_count,
    #                  self._valid_count, self.n_labelled_pixels)
    #     return s

    @property
    def warp_stats(self) -> str:
        return "Warp stats: successful: %i, failed %i, quota: %.1f" %(
            self.n_successful_warp, self.n_failed_warp,
            float(self.n_successful_warp)/(self.n_failed_warp+self.n_successful_warp))

    def warp_cut(
            self,
            inp_src: DataSource,
            target_src: Optional[DataSource],
            warp_prob: Union[float, bool],
            warp_kwargs: Dict[str, Any]
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """
        (Wraps :py:meth:`elektronn3.data.coord_transforms.get_warped_slice()`)

        Cuts a warped slice out of the input and target arrays.
        The same random warping transformation is each applied to both input
        and target.

        Warping is randomly applied with the probability defined by the ``warp_prob``
        parameter (see below).

        Parameters
        ----------
        inp_src: h5py.Dataset
            Input image source (in HDF5)
        target_src: h5py.Dataset
            Target image source (in HDF5)
        warp_prob: float or bool
            False/True disable/enable warping completely.
            If ``warp_prob`` is a float, it is used as the ratio of inputs that
            should be warped.
            E.g. 0.5 means approx. every second call to this function actually
            applies warping to the image-target pair.
        warp_kwargs: dict
            kwargs that are passed through to
            :py:meth:`elektronn2.data.coord_transforms.get_warped_slice()`.
            Can be empty.

        Returns
        -------
        inp: np.ndarray
            (Warped) input image slice
        target_src: np.ndarray
            (Warped) target slice
        """
        if (warp_prob is True) or (warp_prob == 1):  # always warp
            do_warp = True
        elif 0 < warp_prob < 1:  # warp only a fraction of examples
            do_warp = True if (np.random.rand() < warp_prob) else False
        else:  # never warp
            do_warp = False

        if not do_warp:
            warp_kwargs = dict(warp_kwargs)
            warp_kwargs['warp_amount'] = 0

        if target_src is None:
            target_src_shape = None
            target_patch_shape = None
        else:
            target_src_shape = target_src.shape
            target_patch_shape = self.target_patch_shape

        M = coord_transforms.get_warped_coord_transform(
            inp_src_shape=inp_src.shape,
            patch_shape=self.patch_shape,
            aniso_factor=self.aniso_factor,
            target_src_shape=target_src_shape,
            target_patch_shape=target_patch_shape,
            **warp_kwargs
        )

        inp, target = coord_transforms.warp_slice(
            inp_src=inp_src,
            patch_shape=self.patch_shape,
            M=M,
            target_src=target_src,
            target_patch_shape=target_patch_shape,
            target_discrete_ix=self.target_discrete_ix,
            input_discrete_ix=self.input_discrete_ix
        )

        return inp, target

    def _getcube(self) -> Tuple[DataSource, DataSource, int]:
        """
        Draw an example cube according to sampling weight on training data,
        or randomly on valid data
        """
        i = np.random.choice(
            np.arange(len(self.cube_prios)),
            p=self.cube_prios / np.sum(self.cube_prios)
        )
        inp_source = self.inputs[i]
        target_source = None if self.targets is None else self.targets[i]
        return inp_source, target_source, i

    def load_data(self) -> None:
        if len(self.inputs) == len(self.targets) == 0:
            inp_files, target_files = self.open_files()
            self.inputs.extend(inp_files)
            if target_files is None:
                self.targets = None
            else:
                self.targets.extend(target_files)
        else:
            logger.info('Using directly specified data sources.')

        if self.cube_prios is None:
            # If no priorities are given: sample proportionally to target sizes
            #  if available, or else w.r.t. input sizes (voxel counts)
            self.cube_prios = []
            if self.targets is None:
                self.cube_prios = [inp.size for inp in self.inputs]
            else:
                self.cube_prios = [target.size for target in self.targets]
            self.cube_prios = np.array(self.cube_prios, dtype=np.float32) / np.sum(self.cube_prios)

        logger.debug(f'cube_prios = {self.cube_prios}')

    def check_files(self) -> None:
        """
        Check if all files are accessible.
        """
        notfound = False
        give_neuro_data_hint = False
        fullpaths = [f for f, _ in self.input_sources]
        if self.target_sources is not None:
            fullpaths.extend([f for f, _ in self.target_sources])
        for p in fullpaths:
            if not os.path.exists(p):
                print('{} not found.'.format(p))
                notfound = True
                if 'neuro_data_cdhw' in p:
                    give_neuro_data_hint = True
        if give_neuro_data_hint:
            print('\nIt looks like you are referencing the neuro_data_cdhw dataset.\n'
                  'To install the neuro_data_xzy dataset to the default location, run:\n'
                  '  $ wget https://github.com/ELEKTRONN/elektronn.github.io/releases/download/neuro_data_cdhw/neuro_data_cdhw.zip\n'
                  '  $ unzip neuro_data_cdhw.zip -d ~/neuro_data_cdhw')
        if notfound:
            print('\nPlease fetch the necessary dataset and/or '
                  'change the relevant file paths in the network config.')
            sys.stdout.flush()
            sys.exit(1)

    def open_files(self) -> Tuple[List[DataSource], Optional[List[DataSource]]]:
        self.check_files()
        inp_sources, target_sources = [], []
        modestr = 'Training' if self.train else 'Validation'
        memstr = ' (in memory)' if self.in_memory else ''
        logger.info(f'\n{modestr} data set{memstr}:')
        if self.target_sources is None:
            for (inp_fname, inp_key), cube_meta in zip(self.input_sources, self.cube_meta):
                inp_source = HDF5DataSource(fname=inp_fname, key=inp_key, in_memory=self.in_memory)
                logger.info(f'  input:       {inp_fname}[{inp_key}]: {inp_source.shape} ({inp_source.dtype})')
                if not np.all(cube_meta == np.inf):
                    logger.info(f'  cube_meta:   {cube_meta}')
                inp_sources.append(inp_source)
                target_sources = None
        else:
            for (inp_fname, inp_key), (target_fname, target_key), cube_meta in zip(self.input_sources, self.target_sources, self.cube_meta):
                inp_source = HDF5DataSource(fname=inp_fname, key=inp_key, in_memory=self.in_memory)
                target_source = HDF5DataSource(fname=target_fname, key=target_key, in_memory=self.in_memory)
                logger.info(f'  input:       {inp_fname}[{inp_key}]: {inp_source.shape} ({inp_source.dtype})')
                logger.info(f'  with target: {target_fname}[{target_key}]: {target_source.shape} ({target_source.dtype})')
                if not np.all(cube_meta == np.inf):
                    logger.info(f'  cube_meta:   {cube_meta}')
                inp_sources.append(inp_source)
                target_sources.append(target_source)
        logger.info('')

        return inp_sources, target_sources

    def set_offset(self, offset: Sequence[int]) -> None:
        self.offset = np.array(offset)
        self.target_patch_shape = self.patch_shape - self.offset * 2

def get_preview_batch(
        h5data: Tuple[str, str],
        preview_shape: Optional[Tuple[int, ...]] = None,
        transform: Optional[Callable] = None,
        in_memory: bool = False,
        dim: Optional[float] = None,
) -> torch.Tensor:
    fname, key = h5data
    inp_h5 = h5py.File(fname, 'r')[key]
    if in_memory:
        inp_h5 = inp_h5.value
    if dim is None:
        if preview_shape is None:
            raise ValueError('At least one of preview_shape, dim must be defined.')
        dim = len(preview_shape)  # 2D or 3D
    inp_shape = np.array(inp_h5.shape[-dim:])
    if preview_shape is None:  # Slice everything
        inp_lo = np.zeros_like(inp_shape)
        inp_hi = inp_shape
    else:  # Slice only a preview_shape-sized region from the center of the input
        halfshape = np.array(preview_shape) // 2
        inp_center = inp_shape // 2
        inp_lo = inp_center - halfshape
        inp_hi = inp_center + halfshape
        if np.any(inp_center < halfshape):
            raise ValueError(
                'preview_shape is too big for shape of input source.'
                f'Requested {preview_shape}, but can only deliver {tuple(inp_shape)}.'
            )
    memstr = ' (in memory)' if in_memory else ''
    logger.info(f'\nPreview data{memstr}:')
    logger.info(f'  input:       {fname}[{key}]: {inp_h5.shape} ({inp_h5.dtype})\n')
    inp_np = slice_3d(inp_h5, inp_lo, inp_hi, prepend_empty_axis=True)
    if inp_np.ndim == dim + 1:  # Should be dim + 2 for (N, C) dims
        inp_np = inp_np[:, None]  # Add missing C dim
    if transform is not None:
        for n in range(inp_np.shape[0]):  # N is usually 1, so this is only iterated once with n=0
            inp_np[0], _ = transform(inp_np[0], None)
    inp = torch.from_numpy(inp_np)
    return inp


class SimpleNeuroData2d(data.Dataset):
    """ 2D Dataset class for neuro_data_cdhw, reading from a single HDF5 file.

    Delivers 2D image slices from the (H, W) plane at given D indices.
    Not scalable, keeps everything in memory.
    This is just a minimalistic proof of concept.
    """

    def __init__(
            self,
            inp_path=None,
            target_path=None,
            train=True,
            inp_key='raw',
            target_key='lab',
            # offset=(0, 0, 0),
            pool=(1, 1, 1),
            transform: Callable = transforms.Identity(),
            out_channels: Optional[int] = None,
    ):
        super().__init__()
        self.transform = transform
        self.out_channels = out_channels
        cube_id = 0 if train else 2
        if inp_path is None:
            inp_path = expanduser(f'~/neuro_data_cdhw/raw_{cube_id}.h5')
        if target_path is None:
            target_path = expanduser(f'~/neuro_data_cdhw/barrier_int16_{cube_id}.h5')
        self.inp_file = h5py.File(os.path.expanduser(inp_path), 'r')
        self.target_file = h5py.File(os.path.expanduser(target_path), 'r')
        self.inp = self.inp_file[inp_key][()].astype(np.float32)
        self.target = self.target_file[target_key][()].astype(np.int64)
        self.target = self.target[0]  # Squeeze superfluous first dimension
        self.target = self.target[::pool[0], ::pool[1], ::pool[2]]  # Handle pooling (dirty hack TODO)

        # Cut inp and target to same size
        inp_shape = np.array(self.inp.shape[1:])
        target_shape = np.array(self.target.shape)
        diff = inp_shape - target_shape
        offset = diff // 2  # offset from image boundaries

        self.inp = self.inp[
            :,
            offset[0]: inp_shape[0] - offset[0],
            offset[1]: inp_shape[1] - offset[1],
            offset[2]: inp_shape[2] - offset[2],
        ]

        self.close_files()  # Using file contents from memory -> no need to keep the file open.

    def __getitem__(self, index):
        # Get z slices
        inp = self.inp[:, index]
        target = self.target[index]
        inp, target = self.transform(inp, target)
        inp = torch.as_tensor(inp)
        target = torch.as_tensor(target)
        sample = {
            'inp': inp,
            'target': target,
            'cube_meta': np.inf,
            'fname': str(index)
        }
        return sample

    def __len__(self):
        return self.target.shape[0]

    def close_files(self):
        self.inp_file.close()
        self.target_file.close()


# TODO: docs, types
class Segmentation2d(data.Dataset):
    """Simple dataset for 2d segmentation.

    Expects a list of ``input_paths`` and ``target_paths`` where
    ``target_paths[i]`` is the target of ``input_paths[i]`` for all i.
    """
    def __init__(
            self,
            inp_paths,
            target_paths,
            transform=transforms.Identity(),
            offset: Sequence[int] = (0 ,0, 0),
            in_memory=True,
            inp_dtype=np.float32,
            target_dtype=np.int64,
            epoch_multiplier=1,  # Pretend to have more data in one epoch
    ):
        super().__init__()
        self.inp_paths = inp_paths
        self.target_paths = target_paths
        self.transform = transform
        self.offset = offset
        self.in_memory = in_memory
        self.inp_dtype = inp_dtype
        self.target_dtype = target_dtype
        self.epoch_multiplier = epoch_multiplier

        def load_image(fname):
            inp = imageio.imread(fname).astype(np.float32)
            if inp.ndim == 2:
                inp = inp[None]  # (H, W) -> (C=1, H, W)
            elif inp.ndim == 3:
                inp = inp.transpose(2, 0, 1)  # (H, W, C) -> (C, H, W)
            else:
                raise RuntimeError(f'Image {fname} has shape {inp.shape}, but ndim should be 2 or 3.')
            return inp

        if self.in_memory:
            self.inputs = []
            rgb_fnames = {}
            gray_fnames = {}
            for input_path in self.inp_paths:
                if os.path.isdir(input_path):
                    multi_input = []
                    for channel_idx, input_file in enumerate(sorted(glob.glob(str(input_path) + '/*'))):
                        inp = load_image(str(input_file))
                        if inp.shape[0] == 1:
                            gray_fnames[channel_idx] = input_file
                        elif inp.shape[0] == 3:
                            rgb_fnames[channel_idx] = input_file
                        rgb_fname = rgb_fnames.get(channel_idx)
                        if rgb_fname is not None and inp.shape[0] == 1:
                            raise RuntimeError(f'GT input layer {channel_idx} has mixed multi-channel ({rgb_fname}) and single-channel images ({input_file}).')
                        gray_fname = gray_fnames.get(channel_idx)
                        if gray_fname is not None and inp.shape[0] == 3:
                            raise RuntimeError(f'GT input layer {channel_idx} has mixed multi-channel ({input_file}) and single-channel images ({gray_fname}).')
                        multi_input.append(inp)
                    self.inputs.append(np.concatenate(multi_input))
                else:
                    inp = load_image(input_path)
                    if inp.shape[0] == 1:
                        gray_fnames[0] = input_path
                    elif inp.shape[0] == 3:
                        rgb_fnames[0] = input_path
                    if len(rgb_fnames) > 0 and inp.shape[0] == 1 or len(gray_fnames) > 0 and inp.shape[0] == 3:
                        raise RuntimeError(f'Mixed multi-channel ({rgb_fnames[0]}) and single-channel images ({gray_fnames[0]}) in gt.')
                    self.inputs.append(inp)
            self.targets = [
                np.array(imageio.imread(fname)).astype(np.int64)
                for fname in self.target_paths
            ]

    def __getitem__(self, index):
        index %= len(self.inp_paths)  # Wrap around to support epoch_multiplier
        if self.in_memory:
            inp = self.inputs[index]
            target = self.targets[index]
        else:
            fname = self.inp_paths[index]
            inp = imageio.imread(fname).astype(np.float32)
            if inp.ndim == 2:
                inp = inp[None]  # (H, W) -> (C=1, H, W)
            elif inp.ndim == 3:
                inp = inp.transpose(2, 0, 1)  # (H, W, C) -> (C, H, W)
            else:
                raise RuntimeError(f'Image {fname} has shape {inp.shape}, but ndim should be 2 or 3.')
            target = np.array(imageio.imread(self.target_paths[index]), dtype=np.int64)
        while True:  # Only makes sense if RandomCrop is used
            try:
                inp, target = self.transform(inp, target)
                break
            except transforms._DropSample:
                pass
        if np.any(self.offset):
            off = self.offset
            target = target[off[0]:-off[0], off[1]:-off[1]]
        sample = {
            'inp': torch.as_tensor(inp.astype(self.inp_dtype)),
            'target': torch.as_tensor(target.astype(self.target_dtype)),
            'cube_meta': np.inf,
            'fname': str(self.inp_paths[index])
        }
        return sample

    def __len__(self):
        return len(self.target_paths) * self.epoch_multiplier

    def set_offset(self, offset: Sequence[int]) -> None:
        self.offset = offset

# TODO: Document
class Reconstruction2d(data.Dataset):
    """Simple dataset for 2d reconstruction for auto-encoders etc..
    """
    def __init__(
            self,
            inp_paths,
            transform=transforms.Identity(),
            in_memory=True,
            inp_dtype=np.float32,
            epoch_multiplier=1,  # Pretend to have more data in one epoch
    ):
        super().__init__()
        self.inp_paths = inp_paths
        self.transform = transform
        self.in_memory = in_memory
        self.inp_dtype = inp_dtype
        self.epoch_multiplier = epoch_multiplier

        if self.in_memory:
            self.inputs = [
                np.array(imageio.imread(fname)).astype(np.float32)[None]
                for fname in self.inp_paths
            ]

    def __getitem__(self, index):
        index %= len(self.inp_paths)  # Wrap around to support epoch_multiplier
        if self.in_memory:
            inp = self.inputs[index]
        else:
            inp = np.array(imageio.imread(self.inp_paths[index]), dtype=self.inp_dtype)
            if inp.ndim == 2:  # (H, W)
                inp = inp[None]  # (C=1, H, W)
        while True:  # Only makes sense if RandomCrop is used
            try:
                inp, _ = self.transform(inp, None)
                break
            except transforms._DropSample:
                pass
        inp = torch.as_tensor(inp)
        sample = {
            'inp': inp,
            'target': inp,
            'cube_meta': np.inf,
            'fname': str(self.inp_paths[index])
        }
        return sample

    def __len__(self):
        return len(self.inp_paths) * self.epoch_multiplier


class TripletData2d(data.Dataset):
    """Simple dataset for 2D triplet loss training.
    """
    def __init__(
            self,
            inp_paths,
            transform=transforms.Identity(),
            invariant_transform=None,
            in_memory=True,
            inp_dtype=np.float32,
            epoch_multiplier=1,  # Pretend to have more data in one epoch
    ):
        super().__init__()
        self.inp_paths = inp_paths
        self.transform = transform
        self.invariant_transform = invariant_transform
        self.in_memory = in_memory
        self.inp_dtype = inp_dtype
        self.epoch_multiplier = epoch_multiplier

        if self.in_memory:
            self.inputs = [
                np.array(imageio.imread(fname)).astype(np.float32)[None]
                for fname in self.inp_paths
            ]

    def _get(self, index):
        if self.in_memory:
            inp = self.inputs[index]
        else:
            inp = np.array(imageio.imread(self.inp_paths[index]), dtype=self.inp_dtype)
            if inp.ndim == 2:  # (H, W)
                inp = inp[None]  # (C=1, H, W)
        while True:  # Only makes sense if RandomCrop is used
            try:
                inp, _ = self.transform(inp, None)
                break
            except transforms._DropSample:
                pass
        return inp

    def _randidx_excluding(self, exclude):
        while True:
            idx = np.random.randint(0, len(self.inp_paths) // self.epoch_multiplier)
            if idx != exclude:
                return idx

    def __getitem__(self, index):
        index %= len(self.inp_paths)  # Wrap around to support epoch_multiplier
        anchor = self._get(index)
        if self.invariant_transform is None:
            # Assuming a random augmentation transform, the positive image will be different than
            #  the anchor, but it will originate from the same image file.
            #  If random cropping and geometrical transforms are used, make sure that the loss is
            #  not calculated on localized/spatial outputs!
            pos = self._get(index)
        else:
            # Apply an additional transform against which the network should learn invariant behavior
            pos, _ = self.invariant_transform(anchor, None)
        # Sample a negative image from a random different index -> different image
        neg_idx = self._randidx_excluding(index)
        neg = self._get(neg_idx)
        if self.invariant_transform is not None:
            # Also apply the invariant transform to the negative image because otherwise
            #  the model could "cheat" by detecting that the inherent features of this
            #  transform only exist in the positive image.
            neg, _ = self.invariant_transform(neg, None)
        sample = {
            'anchor': torch.as_tensor(anchor),
            'pos': torch.as_tensor(pos),
            'neg': torch.as_tensor(neg),
            'fname': f'ap{index}n{neg_idx}'
        }
        return sample

    def __len__(self):
        return len(self.inp_paths) * self.epoch_multiplier

# TODO: Warn if datasets have no content



import os
import logging
import datetime
from pickle import PicklingError
from typing import Optional, Dict, Tuple, Union, List, Any, Sequence
from sklearn.cluster import KMeans
from sklearn.metrics import v_measure_score
from tqdm import tqdm
import numpy as np
import zipfile
import pprint
from collections import deque
from textwrap import dedent
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("agg")

import inspect
import IPython
import torch
import torch.nn
import torch.utils.data
from torch.utils.tensorboard import SummaryWriter
from torch.utils import collect_env
import torch_geometric
import elektronn3
from elektronn3.training.trainer import _change_log_file_to, NaNException
from elektronn3.training.train_utils import HistoryTracker, Timer, pretty_string_time

logger = logging.getLogger('elektronn3log')


class GNNTrainer:

    tb: SummaryWriter
    terminate: bool
    step: int
    epoch: int
    train_loader: torch.utils.data.DataLoader
    valid_loader: torch.utils.data.DataLoader
    exp_name: str
    save_path: str  # Full path to where training files are stored
    out_channels: Optional[int]  # Number of channels of the network outputs

    def __init__(
        self, 
        model: torch.nn.Module,
        criterion: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        save_root: str,
        data: Tuple[torch_geometric.data.Data],
        valid_metrics: Optional[Dict] = None,
        extra_save_steps: Sequence[int] = (),
        batch_size: int = 1,
        exp_name: Optional[str] = None,
        schedulers: Optional[Dict[Any, Any]] = None,
        early_stopping: Optional[Any] = None,
        num_workers: int = 0,
        enable_tensorboard: bool = True,
        tensorboard_root_path: Optional[str] = None,
        ignore_errors: bool = False,
        ipython_shell: bool = False,
        out_channels: Optional[int] = None,
        seed: Optional[int] = 0,
        mixed_precision: bool = False,
        tqdm_kwargs: Optional[Dict] = None,
    ):
        model.to(device)
        if isinstance(criterion, torch.nn.Module):
            criterion.to(device)

        self.device = device
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.save_root = os.path.expanduser(save_root)
        self.data = data
        self.extra_save_steps = extra_save_steps
        self.batch_size = batch_size
        self.schedulers = schedulers
        self.early_stopping = early_stopping
        self.num_workers = num_workers
        self.ignore_errors = ignore_errors
        self.ipython_shell = ipython_shell
        self.out_channels = out_channels
        self.seed = seed
        self.mixed_precision = mixed_precision
        self.tqdm_kwargs = {} if tqdm_kwargs is None else tqdm_kwargs
        
        self._shell_info = dedent("""
            Entering IPython training shell. To continue, hit Ctrl-D twice.
            To terminate, set self.terminate = True and then hit Ctrl-D twice.
        """).strip()

        lr = self.optimizer.state_dict()["param_groups"][0]["lr"]

        logger.info(f"Training configuration: lr={lr}, criterion={self.criterion.__class__.__name__}, optimizer={self.optimizer.__class__.__name__}, model={self.model.__class__.__name__}, early_stopping={True if self.early_stopping else False}")

        self._tracker = HistoryTracker()
        self._timer = Timer()

        if exp_name is None:
            timestamp = datetime.datetime.now().strftime('%y-%m-%d_%H-%M-%S')
            exp_name = model.__class__.__name__ + '__' + timestamp

        self.exp_name = exp_name
        self.save_path = os.path.join(save_root, self.exp_name)
        if os.path.isdir(self.save_path):
            raise RuntimeError(f"Save path {self.save_path} already exists")

        os.makedirs(self.save_path)
        _change_log_file_to(f'{self.save_path}/elektronn3.log')
        logger.info(f"Writing files to save_path {self.save_path}/\n")

        # Terminate training by setting to true
        self.terminate = False
        self.epoch = 0
        self.patience = self.early_stopping.patience if self.early_stopping else 0
        self._lr_nhood = deque(maxlen=3)
        self.out_channels = out_channels
        self.validate_every_epoch = 1
        self.tb = None

        if enable_tensorboard:
            if tensorboard_root_path is None:
                tb_path = self.save_path
            else:
                tensorboard_root_path = os.path.expanduser(tensorboard_root_path)
                tb_path = os.path.join(tensorboard_root_path, self.exp_name)
                os.makedirs(tb_path, exist_ok=True)
            self.tb = SummaryWriter(tb_path, flush_secs=20)

        self.best_val_loss = np.inf
        self.valid_metrics = {} if valid_metrics is None else valid_metrics
        self.validate = False

    def run(self, max_epochs: int=1, max_runtime=3600*24*7) -> None:
        self.start_time = datetime.datetime.now()
        self.end_time = self.start_time + datetime.timedelta(seconds=max_runtime)
        # Save initial model
        self._save_model(suffix="_initial", verbose=False)
        self._lr_nhood.clear()
        self._lr_nhood.append(self.optimizer.param_groups[0]['lr'])  # LR of the first training step

        stats: Dict[str, Union[float, List[float]]] = {}
        misc: Dict[str, Union[float, List[float]]] = {}
        
        while not self.terminate:
            try:
                timer = Timer()
                
                # Training loss
                dloss = self._train()
                loss = float(dloss.item())

                stats['tr_loss'] = loss
                misc["tr_speed"] = self.data.num_nodes / timer.t_passed
                misc['learning_rate'] = self.optimizer.param_groups[0]['lr']

                self._tracker.update_timeline([self._timer.t_passed, loss])
                
                # Increment epochs
                self._incr_epoch(max_epochs, max_runtime)

                valid_stats = self._validate()
                stats.update(valid_stats)

                self.schedulers['lr'].step(stats['val_loss'])

                # Log to stdout and text log file
                self._log_basic(stats, misc)
                # Render visualizations and log to tensorboard
                self._log_to_tensorboard(stats, misc)

                # Save current model
                self._save_model(val_loss=stats['val_loss'], verbose=False)

                # Save best model
                if stats["val_loss"] < self.best_val_loss:
                    self.best_val_loss = stats["val_loss"]
                    self._save_model(suffix="_best", val_loss=stats["val_loss"])

                if self.early_stopping is not None:
                    self.early_stopping(stats["val_loss"])
                    if self.early_stopping.stop:
                        logger.info(f"Validation loss did not improve for {self.patience} epochs. Terminating...")
                        self.terminate = True 

            except KeyboardInterrupt: 
                if self.ipython_shell:
                    IPython.embed(header=self._shell_info)
                else:
                    break
                if self.terminate:
                    break

            except Exception as e:
                logger.exception("Unhandled exception during training:")
                if self.ignore_errors:
                    pass
                elif self.ipython_shell:
                    print("\nEntering Command line such that Exception can be "
                          "further inspected by user.\n\n")
                    IPython.embed(header=self._shell_info)
                    if self.terminate:
                        break
                else:
                    raise e

        # Save final model
        self._save_model(suffix="_final")
        if self.tb is not None:
            self.tb.close()  # Ensure that everything is flushed

    def _train(self) -> None:
        self.model.train()
        self.optimizer.zero_grad()
        
        x, edge_index, edge_attr, y = self.data.x, self.data.edge_index, self.data.edge_attr, self.data.y
        mask = self.data.train_mask

        if self.model.__class__.__name__ == 'MLP':
            dout = self.model(x[mask])
            dloss = self.criterion(dout, y[mask])
        else:
            attr = edge_attr[:,2]+edge_attr[:,9]
            attr = attr - attr.min()
            min_value = attr.min()
            max_value = attr.max()
            attr.div_(max_value - min_value)
            if self.model.__class__.__name__ == 'GCN':
                dout = self.model(x, edge_index, attr)
            else:
                dout = self.model(x, edge_index, edge_attr[:,[2,9]])

            dloss = self.criterion(dout[mask], y[mask])
        
        if torch.isnan(dloss):
            logger.error(f"NaN loss encountered at step {self.epoch}. Aborting...")
            raise NaNException()
        
        dloss.backward()
        self.optimizer.step()
        
        return dloss

    @torch.no_grad()
    def _validate(self) -> Tuple[Dict[str, float], Dict[str, np.ndarray]]:
        self.model.eval()
        stats = {name: [] for name in self.valid_metrics.keys()}

        # Validation loss
        x, edge_index, edge_attr, y = self.data.x, self.data.edge_index, self.data.edge_attr, self.data.y
        mask = self.data.val_mask

        if self.model.__class__.__name__ == 'MLP':
            dout = self.model(x[mask])
            dloss = self.criterion(dout, y[mask])
            out = dout.detach().cpu().numpy()
        else:
            attr = edge_attr[:,2]+edge_attr[:,9]
            attr = attr - attr.min()
            min_value = attr.min()
            max_value = attr.max()
            attr.div_(max_value - min_value)
            if self.model.__class__.__name__ == 'GCN':
                dout = self.model(x, edge_index, attr)
            else:
                dout = self.model(x, edge_index, edge_attr[:,[2,9]])

            dloss = self.criterion(dout[mask], y[mask])
            out = dout[mask].detach().cpu().numpy()
        
        stats['val_loss'] = float(dloss.item())
        target = y[mask].detach().cpu().numpy()
    
        if self.epoch % self.validate_every_epoch == 0:
            
            for name, evaluator in self.valid_metrics.items():
                stats[name] = evaluator(target, out)

        self.model.train()

        return stats

    def _incr_epoch(self, max_epochs, max_runtime):
        self.epoch += 1
        
        if self.epoch >= max_epochs:
            logger.info(f"max_epochs ({max_epochs}) exceeded. Terminating...")
            self.terminate = True

        if datetime.datetime.now() >= self.end_time:
            logger.info(f"max_runtime ({max_runtime} seconds) exceeded. Terminating...")
            self.terminate = True

    def _scheduler_step(self, loss):
        """Update schedules"""
        for sched in self.schedulers.values():
            # support ReduceLROnPlateau; doc. uses validation loss instead
            # http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau
            if 'metrics' in inspect.signature(sched.step).parameters:
                sched.step(metrics=loss)
            else:
                sched.step()
        # Append LR of the next iteration (after sched.step()) for local LR minima detection
        self._lr_nhood.append(self.optimizer.param_groups[0]['lr'])

    def _save_model(
        self, 
        suffix: str="",
        verbose: bool = True,
        val_loss=np.nan
    ) -> None:
        log = logger.info if verbose else logger.debug

        model = self.model
        model_trainmode = model.training

        state_dict_path = os.path.join(self.save_path, f"state_dict{suffix}.pth")
        model_path = os.path.join(self.save_path, f"model{suffix}.pt")

        info = {
            'epoch': self.epoch,
            'best_val_loss': self.best_val_loss,
            'val_loss': val_loss,
            'elektronn3.__version__': elektronn3.__version__,
            'env_info': collect_env.get_pretty_env_info() 
        }

        info = {k: str(v) for k, v in info.items()}

        torch.save({
            'model_state_dict': model.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'info': info
        }, state_dict_path)

        log(f'Saved state_dict as {state_dict_path}')
        pts_model_path = f"{model_path}"

        try:
            torch.save(model, model_path)
            log(f'Saved model as {model_path}')
        
        except (TypeError, PicklingError) as exc:
            raise exc

        finally:
            model.training = model_trainmode
        
        if os.path.isfile(pts_model_path):
            with zipfile.ZipFile(pts_model_path, 'a', compression=zipfile.ZIP_DEFLATED) as zfile:
                infostr = pprint.pformat(info, indent=2, width=120)
                zfile.writestr('info.txt', infostr)

    def _log_basic(self, stats, misc):
        tr_loss = stats['tr_loss']
        val_loss = stats['val_loss']
        lr = misc['learning_rate']
        tr_speed = misc['tr_speed']
        t = pretty_string_time(self._timer.t_passed)
        text = f'epoch={self.epoch:04d}, tr_loss={tr_loss:.3f}, val_loss={val_loss:.3f}, '
        text += f'lr={lr:.2e}, tr_speed={tr_speed:.2f} it/s, {t}'
        logger.info(text)

    def _log_to_tensorboard(
        self,
        stats: Dict,
        misc: Dict,
        file_stats: Optional[Dict] = None
    ) -> None:
        if self.tb:
            try:
                self._tb_log_scalars(stats, "stats")
                self._tb_log_scalars(misc, "misc")
                if file_stats is not None:
                    self._tb_log_scalars(file_stats, "file_stats")
                self._tb_log_histograms()
            except Exception as exc:
                logger.exception('Error occured while logging to tensorboard:')
                raise exc

    def _tb_log_scalars(
            self,
            scalars: Dict[str, float],
            tag: str = 'default'
    ) -> None:
        for key, value in scalars.items():
            if isinstance(value, (list, tuple, np.ndarray)):
                for i in range(len(value)):
                    if not np.isnan(value[i]):
                        self.tb.add_scalar(f'{tag}/{key}', value[i], self.epoch - len(value) + i)
            elif not np.isnan(value):
                self.tb.add_scalar(f'{tag}/{key}', value, self.epoch)

    def _tb_log_histograms(self) -> None:
        """Log histograms of model parameters and their current gradients.
        Make sure to run this between ``backward()`` and ``zero_grad()``,
        because otherwise gradient histograms will only consist of zeros.
        """
        for name, param in self.model.named_parameters():
            self.tb.add_histogram(f'param/{name}', param, self.epoch)
            grad = param.grad if param.grad is not None else torch.tensor(0)
            self.tb.add_histogram(f'grad/{name}', grad, self.epoch)


import collections
import os
import random
from typing import Callable, Dict, Optional, Sequence

import knossos_utils
import numpy as np
import torch
import torch.utils.data
from elektronn3.data import transforms
from elektronn3.data.knossos import KnossosRawData


class KnossosLabels(torch.utils.data.Dataset):
    """Delivers label and raw data as patches that are randomly
        sampled from a KNOSSOS dataset. The labels are extracted from a kzip
        file and the corresponding raw data associated with them is returned.
        Supports 2D and 3D (choice is determined from the length of the
        ``patch_shape`` param).

        Args:
            conf_path_label: Path to KNOSSOS .conf file corresponding to the labels
            conf_path_raw_data: Path to KNOSSOS .conf file corresponding to the raw data
            dir_path_label: Directory containing label kzip files
            patch_shape: Shape (zyx) of patches that should be sampled from the
                dataset.
            transform: Transformation to be applied to the loaded data.
                See :py:mod:`elektronn3.data.transforms`.
            mag: KNOSSOS magnification number
            epoch_size: Determines the length (``__len__``) of the ``Dataset``
                iterator. ``epoch_size`` can be set to an arbitrary value and
                doesn't have any effect on the content of produced training
                samples.
            label_names: A sequence of labels to be extracted. Extraction of label data is
                only done on files of the form ``label_names.k.zip``. If``None``, label
                data from all the kzip files is extracted. Useful when each kzip contains data
                for only one label.
            knossos_bounds: List of Tuple of boundary coordinates (xyz) that constrain the region
            of where patches should be sampled from within the KNOSSOS dataset.
            E.g. ``knossos_bounds=[((256, 256, 128), (512, 512, 512))]`` means that only
            the region between the low corner (x=256, y=256, z=128) and the high corner
            (x=512, y=512, z=512) of the dataset is considered. The bounds should be one of the
            bounds of different patches in the knossos dataset. If ``None``, then whole dataset is returned.
            label_offset: offset of the indices in the knossos labels, default is 0.
            label_order: change the order of the classes in the labels
           """

    def __init__(
            self,
            conf_path_label: str,
            conf_path_raw_data: str,
            dir_path_label: str,
            patch_shape: Sequence[int],  # [z]yx
            transform: Callable = transforms.Identity(),
            mag: int = 1,
            epoch_size: int = 100,
            label_names: Optional[Sequence[str]] = None,
            knossos_bounds: Optional[Sequence[Sequence[Sequence[int]]]] = None,  # xyz
            label_offset: int = 0,
            label_order: Optional[Sequence[int]] = None
    ):
        self.conf_path_label = conf_path_label
        self.conf_path_raw_data = conf_path_raw_data
        self.patch_shape = np.array(patch_shape)
        self.dim = len(self.patch_shape)
        patch_shape_xyz = self.patch_shape[::-1]  # zyx -> xyz
        if self.dim == 2:
            patch_shape_xyz = np.array([*patch_shape_xyz, 1])  # z=1 for 2D
        self.patch_shape_xyz = patch_shape_xyz
        self.transform = transform
        self.mag = mag
        self.epoch_size = epoch_size
        self.kd = knossos_utils.KnossosDataset(self.conf_path_label, show_progress=False)
        self.inp_targets = []
        self.file_bounds = {}
        self.kzip_files_path = []
        self.dir_path = dir_path_label
        self.knossos_bounds = knossos_bounds
        self.label_offset = label_offset  # todo: verify correct handling of this offset
        self.label_order = label_order

        self._get_file_bounds(label_names)
        self._get_data()

    def _get_file_bounds(self, label_names):
        if self.dir_path:
            if label_names is None:
                self.kzip_files_path = [f for f in os.listdir(self.dir_path) if f.endswith('.k.zip')]
            else:
                for name in label_names:
                    self.kzip_files_path += [f for f in os.listdir(self.dir_path) if f.endswith(name + '.k.zip')]

            for file in self.kzip_files_path:
                zip_path = self.dir_path + "/" + file
                bounds = self.kd.get_movement_area(zip_path)
                offset = tuple(bounds[0])
                max_values = tuple(bounds[1])
                if (offset, max_values) not in self.file_bounds.keys():
                    self.file_bounds[(offset, max_values)] = [zip_path]
                else:
                    self.file_bounds[(offset, max_values)].append(zip_path)

            for offset, max_values in self.file_bounds.keys():
                if len(self.file_bounds[offset, max_values]) != len(label_names):
                    print(f"Warning: Found {len(self.file_bounds[offset, max_values])} label files for "
                          f"offset {offset} and max_values {max_values}, but expected {len(label_names)}")
        else:
            raise OSError(f"Path to conf directory is None!")

    def _get_data(self):
        voxels_per_patch = []
        for bounds, paths in self.file_bounds.items():
            if self.knossos_bounds is None or bounds in self.knossos_bounds:
                size = tuple(np.array(bounds[1]) - np.array(bounds[0]))
                labels_patch = np.zeros(size[::-1])  # the labels of the current patch
                non_background_label_locations = np.zeros_like(labels_patch)
                for zip_path in paths:
                    data = self.kd._load_kzip_seg(zip_path, bounds[0], size,
                                                  self.mag,
                                                  padding=0,
                                                  apply_mergelist=False,
                                                  return_dataset_cube_if_nonexistent=False,
                                                  expand_area_to_mag=False)
                    labels_patch += data
                    non_background_label_locations += data != 0

                overlapping_locations = non_background_label_locations > 1

                if overlapping_locations.sum() != 0:
                    print(
                        f"Detected {overlapping_locations.sum()} overlapping/contradicting labels for "
                        f"labels with bounds {bounds}")
                    # todo: better assign one of the labels instead of background?
                    labels_patch[overlapping_locations] = 0

                # todo: maybe just create one KnossosRawData instance with no bounds, then sample directly?
                inp_raw_data = KnossosRawData(conf_path=self.conf_path_raw_data,
                                              patch_shape=self.patch_shape,
                                              bounds=np.array(bounds) + self.label_offset,
                                              mag=self.mag, mode='disk', epoch_size=self.epoch_size,
                                              disable_memory_check=False, verbose=False)  # xyz form
                target = {'data': labels_patch, 'min_bound': bounds[0],
                          'max_bound': bounds[1], 'fname': paths}  # zyx form
                self.inp_targets.append((inp_raw_data, target))
                voxels_per_patch.append(labels_patch.size)

        self.patch_weights = np.array(voxels_per_patch) / sum(voxels_per_patch)
        print("patch weights:", self.patch_weights)

    def __getitem__(self, index: int) -> Dict[str, torch.Tensor]:
        idx = np.random.choice(np.arange(start=0, stop=len(self.patch_weights)), p=self.patch_weights)
        inp_raw_data, target = self.inp_targets[idx]
        raw_dict = inp_raw_data.__getitem__(0)
        inp = raw_dict['inp']  # zyx
        random_offset = raw_dict["offset"] - self.label_offset  # xyz

        min_indices = random_offset[::-1] - target['min_bound'][::-1]  # computes indices in labeled region
        max_indices = min_indices + self.patch_shape_xyz[::-1]
        label = target['data'][tuple(slice(*i) for i in zip(min_indices,
                                                            max_indices))]
        if self.dim == 2:
            label = label[0]

        inp, label = self.transform(inp.data.numpy(), label)

        sample = {
            'inp': torch.as_tensor(inp),
            'target': torch.as_tensor(label).long(),
            'fname': target['fname']
        }
        if self.label_order is not None:
            sample_target = sample['target'].detach().clone()
            for i, label in enumerate(self.label_order):
                sample['target'][sample_target == i] = label
        return sample

    def __len__(self) -> int:
        return self.epoch_size



from torch import nn
import torch
from torch.nn import functional as F

# No useful architectures here, just some small test networks.


class Simple3DNet(nn.Module):
    def __init__(self, n_out_channels=2):
        super().__init__()
        self.n_out_channels = n_out_channels
        self.conv = nn.Sequential(
            nn.Conv3d(1, 10, 3, padding=1), nn.ReLU(),
            nn.Conv3d(10, 10, 3, padding=1), nn.ReLU(),
            nn.Conv3d(10, n_out_channels, 1)
        )

    def forward(self, x):
        x = self.conv(x)
        return x


class Extended3DNet(nn.Module):
    def __init__(self, n_out_channels=2):
        super().__init__()
        self.n_out_channels = n_out_channels
        self.conv = nn.Sequential(
            nn.Conv3d(1, 64, 5, padding=2), nn.ReLU(),
            nn.Conv3d(64, 64, 5, padding=2), nn.ReLU(),
            nn.MaxPool3d(2),
            nn.Conv3d(64, 64, 3, padding=2), nn.ReLU(),
            nn.Conv3d(64, 64, 3, padding=1), nn.ReLU(),
            nn.Conv3d(64, 64, 3, padding=0), nn.ReLU(),
            nn.Conv3d(64, n_out_channels, 1),
        )

    def forward(self, x):
        original_size = x.size()[2:]
        x = self.conv(x)
        x = F.upsample(x, original_size)
        return x


class N3DNet(nn.Module):
    def __init__(self, n_out_channels=2):
        super().__init__()
        self.n_out_channels = n_out_channels
        self.neuro3d_seq = nn.Sequential(
            nn.Conv3d(1, 20, (1,5,5), padding=(0,2,2)), nn.ReLU(),
            nn.Conv3d(20, 30, (1,5,5), padding=(0,2,2)), nn.ReLU(),
            nn.MaxPool3d((2,2,2)),
            nn.Conv3d(30, 40, (1,5,5), padding=(0,2,2)), nn.ReLU(),
            nn.Conv3d(40, 80, (3,3,3), padding=(1,1,1)), nn.ReLU(),
            nn.Conv3d(80, 100, (3,3,3), padding=(1,1,1)), nn.ReLU(),
            nn.Conv3d(100, 150, (1,3,3), padding=(0,1,1)), nn.ReLU(),
            nn.Conv3d(150, 50, (1,1,1)), nn.ReLU(),
            nn.Conv3d(50, n_out_channels, (1,1,1))
        )

    def forward(self, x):
        original_size = x.size()[2:]
        x = self.neuro3d_seq(x)
        x = F.upsample(x, original_size)
        return x


class Conv3DLayer(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size,
                 batch_norm=True, pooling=None, dropout_rate=None,
                 act=None):
        super().__init__()
        if act is None:
            act = nn.ReLU()
        seq = [nn.Conv3d(in_channels, out_channels, kernel_size)]
        if batch_norm:
            seq += [nn.BatchNorm3d(out_channels)]
        seq += [act]
        if pooling is not None:
            seq += [nn.MaxPool3d(pooling)]
        if dropout_rate is not None:
            seq += [nn.Dropout3d(dropout_rate)]
        self.conv3_layer = nn.Sequential(*seq)

    def forward(self, x):
        return self.conv3_layer(x)


class StackedConv2Scalar(nn.Module):
    def __init__(self, in_channels, n_classes, dropout_rate=0.05, act='relu'):
        super().__init__()
        if act == 'relu':
            act = nn.ReLU()
        elif act == 'leaky_relu':
            act = nn.LeakyReLU()
        self.seq = nn.Sequential(
            Conv3DLayer(in_channels, 20, (1, 5, 5), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(20, 30, (1, 5, 5), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(30, 40, (1, 4, 4), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(40, 50, (1, 4, 4), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(50, 60, (1, 2, 2), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(60, 70, (1, 1, 1), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(70, 70, (1, 1, 1), pooling=(1, 1, 1),
                        dropout_rate=dropout_rate, act=act),
        )
        self.adaptavgpool = nn.AdaptiveAvgPool1d(100)
        self.fc = nn.Sequential(
            nn.Linear(100, 50),
            act,
            nn.Linear(50, 30),
            act,
            nn.Linear(30, n_classes),
        )

    def forward(self, x):
        x = self.seq(x)
        x = x.view(x.size()[0], 1, -1)  # AdaptiveAvgPool1d requires input of shape B C D
        x = self.adaptavgpool(x)
        x = self.fc(x.squeeze(1))  # remove auxiliary axis -> B C with C = n_classes
        return x


class StackedConv2ScalarWithLatentAdd(nn.Module):
    def __init__(self, in_channels, n_classes, dropout_rate=0.05, act='relu',
                 n_scalar=1):
        super().__init__()
        if act == 'relu':
            act = nn.ReLU()
        elif act == 'leaky_relu':
            act = nn.LeakyReLU()
        self.seq = nn.Sequential(
            Conv3DLayer(in_channels, 20, (1, 5, 5), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(20, 30, (1, 5, 5), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(30, 40, (1, 4, 4), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(40, 50, (1, 4, 4), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(50, 60, (1, 2, 2), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(60, 70, (1, 1, 1), pooling=(1, 2, 2),
                        dropout_rate=dropout_rate, act=act),
            Conv3DLayer(70, 70, (1, 1, 1), pooling=(1, 1, 1),
                        dropout_rate=dropout_rate, act=act),
        )
        self.adaptavgpool = nn.AdaptiveAvgPool1d(100)
        self.fc = nn.Sequential(
            nn.Linear(100 + n_scalar, 50),
            act,
            nn.Linear(50, 30),
            act,
            nn.Linear(30, n_classes),
        )

    def forward(self, x, scal):
        x = self.seq(x)
        x = x.view(x.size()[0], 1, -1)  # AdaptiveAvgPool1d requires input of shape B C D
        x = self.adaptavgpool(x).squeeze(1)
        x = torch.cat((x, scal), 1)
        x = self.fc(x)  # remove auxiliary axis -> B C with C = n_classes
        return x



# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Martin Drawitsch, Philipp Schubert, Marius Killinger


import logging
import os
import signal
from typing import Dict, Sequence, Tuple

import h5py
import numpy as np
import torch

from elektronn3 import floatX
from elektronn3.data.sources import DataSource

logger = logging.getLogger("elektronn3log")


eps = 0.0001  # To avoid divisions by zero


def _to_full_numpy(seq) -> np.ndarray:
    if isinstance(seq, np.ndarray):
        return seq
    elif isinstance(seq, DataSource):
        return seq[()]
    elif isinstance(seq[0], np.ndarray):
        return np.array(seq)
    elif isinstance(seq[0], h5py.Dataset) or isinstance(seq[0], DataSource):
        # Explicitly pre-load all dataset values into ndarray format
        return np.array([x[()] for x in seq])
    else:
        raise ValueError('inputs must be an ndarray, a sequence of ndarrays '
                         'or a sequence of h5py.Datasets.')


def calculate_means(inputs: Sequence) -> Tuple[float]:
    inputs = [
        _to_full_numpy(inp).reshape(inp.shape[0], -1)  # Flatten every dim except C
        for inp in inputs
    ]  # Necessary if shapes don't match
    # Preserve C, but concatenate everything else into one flat dimension
    inputs = np.concatenate(inputs, axis=1)
    means = np.mean(inputs, axis=1)
    return tuple(means)


def calculate_stds(inputs: Sequence) -> Tuple[float]:
    inputs = [
        _to_full_numpy(inp).reshape(inp.shape[0], -1)  # Flatten every dim except C
        for inp in inputs
    ]  # Necessary if shapes don't match
    # Preserve C, but concatenate everything else into one flat dimension
    inputs = np.concatenate(inputs, axis=1)
    stds = np.std(inputs, axis=1)
    return tuple(stds)


def calculate_offset(model, tile_shape=None):
    with torch.no_grad():
        param = next(model.parameters())
        if tile_shape:
            example_inp = torch.randn(1, param.size()[1], *tile_shape, device=param.device, dtype=param.dtype)
            example_out = model.eval().forward(example_inp)
        else:
            try: # default for 3d UNet
                example_inp = torch.randn(1, param.size()[1], 90, 90, 90, device=param.device, dtype=param.dtype)
                example_out = model.eval().forward(example_inp)
            except RuntimeError: # default for 2d UNet
                example_inp = torch.randn(1, param.size()[1], 186, 186, device=param.device, dtype=param.dtype)
                example_out = model.eval().forward(example_inp)
    offset = np.subtract(example_inp.shape[2:], example_out.shape[2:]) // 2
    logger.info(f'Inferred target offset: {offset}.')
    return offset


def get_class_counts(targets: Sequence[np.ndarray]) -> Dict[int, str]:
    """Get a dict that maps each target class to its number of labeled pixels/voxels"""
    targets = np.concatenate([
        _to_full_numpy(target).flatten()  # Flatten every dim except C
        for target in targets
    ])  # Necessary if shapes don't match
    total = targets.size
    classes, counts = np.unique(targets, return_counts=True)
    cc = {cl: f'{co} ({co / total * 100:.2f}%)' for cl, co in zip(classes, counts)}
    return cc


def get_nonzero_label_ratio(targets: Sequence[np.ndarray]) -> float:
    targets = np.concatenate([
        _to_full_numpy(target).flatten()  # Flatten every dim except C
        for target in targets
    ])  # Necessary if shapes don't match
    num_nonzero = np.count_nonzero(targets)
    return num_nonzero / targets.size


def calculate_class_weights(
        targets: Sequence[np.ndarray],
        mode='inverse'
) -> np.ndarray:
    """Calulate class weights that assign more weight to less common classes.

    The weights can then be used for loss function rebalancing (e.g. for
    CrossEntropyLoss it's very important to do this when training on
    datasets with high class imbalance."""

    targets = np.concatenate([
        _to_full_numpy(target).flatten()  # Flatten every dim except C
        for target in targets
    ])  # Necessary if shapes don't match

    def __inverse(targets):
        """The weight of each class c1, c2, c3, ... with labeled-element
        counts n1, n2, n3, ... is assigned by weight[i] = N / n[i],
        where N is the total number of all elements in all ``targets``.
        (We could achieve the same relative weight proportions by
        using weight[i] = 1 / n[i], as proposed in
        https://arxiv.org/abs/1707.03237, but we multiply by N to prevent
        very small values that could lead to numerical issues."""
        classes = np.arange(0, targets.max() + 1)
        # Count total number of labeled elements per class
        num_labeled = np.array([
            np.sum(np.equal(targets, c))
            for c in classes
        ], dtype=np.float32)
        class_weights = (targets.size / (num_labeled + eps)).astype(np.float32)
        return class_weights

    def __norpf_inverse(targets):
        classes = np.arange(0, targets.max() + 1)
        # Count total number of labeled elements per class
        num_labeled = np.array([
            np.sum(np.equal(targets, c))
            for c in classes
        ], dtype=np.float32)
        class_weights = (num_labeled / targets.size).astype(np.float32)
        return class_weights

    def __binmean(targets):
        """Use the mean of the targets to determine class weights.

        This assumes a binary segmentation problem (background/foreground) and
        breaks in a multi-class setting."""
        target_mean = np.mean(targets)
        bg_weight = target_mean / (1. + target_mean)
        fg_weight = 1. - bg_weight
        # class_weights = torch.tensor([bg_weight, fg_weight])
        class_weights = np.array([bg_weight, fg_weight], dtype=np.float32)
        return class_weights

    if mode == 'inverse':
        return __inverse(targets)
    elif mode == 'inversesquared':
        return __inverse(targets) ** 2
    elif mode == 'norpf_inverse':
        return __norpf_inverse(targets)
    elif mode == 'binmean':
        return __binmean(targets)


def calculate_nd_slice(src, coords_lo, coords_hi):
    """Calculate the ``slice`` object list that is used as indices for
    reading from a data source.

    Unfortunately, this kind of slice list is not yet supported by h5py.
    It only works with numpy arrays."""
    # Separate spatial dimensions (..., H, W) from nonspatial dimensions (C, ...)
    spatial_dims = len(coords_lo)  # Assuming coords_lo addresses all spatial dims
    nonspatial_dims = src.ndim - spatial_dims  # Assuming every other dim is nonspatial

    # Calculate necessary slice indices for reading the file
    nonspatial_slice = [  # Slicing all available content in these dims.
        slice(0, src.shape[i]) for i in range(nonspatial_dims)
    ]
    spatial_slice = [  # Slice only the content within the coordinate bounds
        slice(coords_lo[i], coords_hi[i]) for i in range(spatial_dims)
    ]
    full_slice = nonspatial_slice + spatial_slice
    return full_slice


def save_to_h5(data, path, hdf5_names=None, overwrite=False, compression=True):
    """
    Saves data to HDF5 File.

    Parameters
    ----------
    data: list or dict of np.arrays
        if list, hdf5_names has to be set.
    path: str
        forward-slash separated path to file
    hdf5_names: list of str
        has to be the same length as data
    overwrite : bool
        determines whether existing files are overwritten
    compression : bool
        True: compression='gzip' is used which is recommended for sparse and
        ordered data

    Returns
    -------
    nothing

    """
    if (not type(data) is dict) and hdf5_names is None:
        raise Exception("hdf5names has to be set if data is a list")
    if os.path.isfile(path) and overwrite:
        os.remove(path)
    f = h5py.File(path, "w")
    if type(data) is dict:
        for key in data.keys():
            if compression:
                f.create_dataset(key, data=data[key], compression="gzip")
            else:
                f.create_dataset(key, data=data[key])
    else:
        if len(hdf5_names) != len(data):
            f.close()
            raise Exception("Not enough or to much hdf5-names given!")
        for nb_data in range(len(data)):
            if compression:
                f.create_dataset(hdf5_names[nb_data], data=data[nb_data],
                                 compression="gzip")
            else:
                f.create_dataset(hdf5_names[nb_data], data=data[nb_data])
    f.close()


def as_floatX(x):
    if not hasattr(x, '__len__'):
        return np.array(x, dtype=floatX)
    return np.ascontiguousarray(x, dtype=floatX)


def squash01(img: np.ndarray) -> np.ndarray:
    """Squash image array to the value range [0, 1] (no clipping).

    This can be used to prepare network outputs or normalized inputs
    for plotting and generic image processing functions.
    """
    img = img.astype(np.float32)
    squashed = (img - np.min(img)) / np.ptp(img)
    return squashed


# https://gist.github.com/tcwalther/ae058c64d5d9078a9f333913718bba95
# class based on: http://stackoverflow.com/a/21919644/487556
class DelayedInterrupt:
    def __init__(self, signals):
        if not isinstance(signals, list) and not isinstance(signals, tuple):
            signals = [signals]
        self.sigs = signals

    def __enter__(self):
        self.signal_received = {}
        self.old_handlers = {}
        for sig in self.sigs:
            self.signal_received[sig] = False
            self.old_handlers[sig] = signal.getsignal(sig)
            def handler(s, frame):
                logger.warning('Signal %s received. Delaying KeyboardInterrupt.' % sig)
                self.signal_received[sig] = (s, frame)
                # Note: in Python 3.5, you can use signal.Signals(sig).name
            self.old_handlers[sig] = signal.getsignal(sig)
            signal.signal(sig, handler)

    def __exit__(self, type, value, traceback):
        for sig in self.sigs:
            signal.signal(sig, self.old_handlers[sig])
            if self.signal_received[sig] and self.old_handlers[sig]:
                self.old_handlers[sig](*self.signal_received[sig])


class CleanExit:
    # https://stackoverflow.com/questions/4205317/capture-keyboardinterrupt-in-python-without-try-except
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_type is KeyboardInterrupt:
            logger.warning('Delaying KeyboardInterrupt.')
            return True
        return exc_type is None


class GracefulInterrupt:
    # by https://stackoverflow.com/questions/18499497/how-to-process-sigterm-signal-gracefully
    now = False

    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, sig, frame):
        logger.warning('Signal %s received. Delaying KeyboardInterrupt.' % sig)
        self.now = True



import os
import logging
import datetime
from pickle import PicklingError
from typing import Optional, Dict, Tuple, Union, List, Any, Sequence
from tqdm import tqdm
import numpy as np
import zipfile
import pprint
import matplotlib
import matplotlib.pyplot as plt
from collections import deque
from textwrap import dedent
matplotlib.use("agg")

import inspect
import IPython
import torch
import torch.nn
import torch.nn.functional as F
import torch.utils.data
from torch.utils.tensorboard import SummaryWriter
from torch.utils import collect_env
from torch.cuda import amp
import torch_geometric
from torch_geometric.loader import NeighborLoader, ImbalancedSampler
from torch_geometric.data import Data
from torch_geometric.utils import degree, dropout_adj
from sklearn.cluster import KMeans, MiniBatchKMeans
from sklearn.linear_model import SGDClassifier
from sklearn.metrics.cluster import homogeneity_completeness_v_measure
# import tensorboardX

import elektronn3
from elektronn3.training.trainer import _change_log_file_to, NaNException
from elektronn3.training.train_utils import HistoryTracker, Timer, pretty_string_time


logger = logging.getLogger('elektronn3log')

def visualize_embeddings(x, color, epoch):
    mapping = {0: "MSN", 1: "LMAN", 2: "HVC"}

    fig, ax = plt.subplots(figsize=(10, 10))
    scatter = ax.scatter(x[:, 0], x[:, 1], c=color, s=15)
    handles, labels = scatter.legend_elements()
    labels = [mapping[int(l.split("{")[1][0])] for l in labels]
    ax.set_xticks([])
    ax.set_yticks([])
    ax.legend(handles, labels, loc="best", title="Cell types")
    ax.set_title(f"Node embeddings (t-SNE), Epoch {epoch}")

    # convert figure to numpy array
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep="")
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))

    return data.reshape(3, data.shape[0], data.shape[1])  # (3, height, width)


class GNNTrainer:

    tb: SummaryWriter
    terminate: bool
    step: int
    epoch: int
    train_loader: torch.utils.data.DataLoader
    valid_loader: torch.utils.data.DataLoader
    exp_name: str
    save_path: str  # Full path to where training files are stored
    out_channels: Optional[int]  # Number of channels of the network outputs

    def __init__(
        self, 
        model: torch.nn.Module,
        criterion: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        save_root: str,
        inductive: bool,
        data: torch_geometric.data.Data,
        num_neighbors: Optional[List[int]] = [25, 10],
        edge_attr_ix: Optional[Tuple[int]] = None,
        valid_metrics: Optional[Dict] = None,
        extra_save_steps: Sequence[int] = (),
        batch_size: int = 1,
        exp_name: Optional[str] = None,
        schedulers: Optional[Dict[Any, Any]] = None,
        early_stopping: Optional[Any] = None,
        self_supervised: Optional[bool] = False,
        embed_loss: Optional[bool] = False,
        negative_sampling: Optional[bool] = False,
        variational: Optional[bool] = False,
        num_workers: int = 0,
        enable_tensorboard: bool = True,
        tensorboard_root_path: Optional[str] = None,
        ignore_errors: bool = False,
        ipython_shell: bool = False,
        out_channels: Optional[int] = None,
        seed: Optional[int] = 0,
        debug_mode: bool = False,
        mixed_precision: bool = False,
        tqdm_kwargs: Optional[Dict] = None,
    ):
        model.to(device)
        if isinstance(criterion, torch.nn.Module):
            criterion.to(device)

        self.device = device
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.extra_save_steps = extra_save_steps
        self.save_root = os.path.expanduser(save_root)
        self.inductive = inductive
        self.data = data
        self.num_neighbors = num_neighbors
        self.edge_attr_ix = edge_attr_ix
        self.batch_size = batch_size
        self.schedulers = schedulers
        self.early_stopping = early_stopping
        self.self_supervised = self_supervised
        self.embed_loss = embed_loss
        self.negative_sampling = negative_sampling
        self.variational = variational
        self.num_workers = num_workers
        self.ignore_errors = ignore_errors
        self.ipython_shell = ipython_shell
        self.out_channels = out_channels
        self.seed = seed
        self.debug_mode = debug_mode
        self.mixed_precision = mixed_precision
        self.tqdm_kwargs = {} if tqdm_kwargs is None else tqdm_kwargs

        self._shell_info = dedent("""
            Entering IPython training shell. To continue, hit Ctrl-D twice.
            To terminate, set self.terminate = True and then hit Ctrl-D twice.
        """).strip()
        
        lr = self.optimizer.state_dict()["param_groups"][0]["lr"]
        logger.info(f"Training configuration: lr={lr}, batchsize={batch_size}, criterion={self.criterion.__class__.__name__}, optimizer={self.optimizer.__class__.__name__}, model={self.model.__class__.__name__}, num_neighbors={num_neighbors}, early_stopping={True if self.early_stopping else False}")

        self._tracker = HistoryTracker()
        self._timer = Timer()

        if exp_name is None:
            timestamp = datetime.datetime.now().strftime('%y-%m-%d_%H-%M-%S')
            exp_name = model.__class__.__name__ + '__' + timestamp

        self.exp_name = exp_name
        self.save_path = os.path.join(save_root, self.exp_name)
        if os.path.isdir(self.save_path):
            raise RuntimeError(f"Save path {self.save_path} already exists")

        os.makedirs(self.save_path)
        _change_log_file_to(f'{self.save_path}/elektronn3.log')
        logger.info(f"Writing files to save_path {self.save_path}/\n")

        # Terminate training by setting to true
        self.terminate = False
        self.step = 0
        self.epoch = 0
        self.patience = self.early_stopping.patience if self.early_stopping else 0
        self._lr_nhood = deque(maxlen=3)
        self.validate_every_epoch = 5

        if self.debug_mode:
            self.validate_every_epoch = 1

        self.tb = None

        if enable_tensorboard:
            if tensorboard_root_path is None:
                tb_path = self.save_path
            else:
                tensorboard_root_path = os.path.expanduser(tensorboard_root_path)
                tb_path = os.path.join(tensorboard_root_path, self.exp_name)
                os.makedirs(tb_path, exist_ok=True)
            self.tb = SummaryWriter(tb_path, flush_secs=20)

        if self.debug_mode:
            # select a small subset of data for debugging purposes
            mask = torch.rand(data.num_nodes) > 0.9
            data = data.subgraph(mask)
            self.validate_every_epoch = 1  # validate in the first epoch

        if inductive:
            logger.info("Using inductive training")
            train_data = data.subgraph(data.train_mask)
            val_data = data.subgraph(data.val_mask)
            # test_data = data.subgraph(data.test_mask)

            self.train_loader = NeighborLoader(
                train_data,
                num_neighbors=self.num_neighbors,
                shuffle=True,
                batch_size=batch_size,
            )
            
            self.valid_loader = NeighborLoader(
                val_data,
                num_neighbors=self.num_neighbors,
                shuffle=False,
                batch_size=batch_size
            )
        
        else: 
            logger.info("Using transductive training")
            sampler = ImbalancedSampler(data, input_nodes=data.train_mask)
            self.train_loader = NeighborLoader(
                data,
                num_neighbors=self.num_neighbors,
                input_nodes=data.train_mask,
                # shuffle=True,
                batch_size=batch_size,
                sampler=sampler
            )

            self.valid_loader = NeighborLoader(
                data,
                num_neighbors=self.num_neighbors,
                input_nodes=data.val_mask,
                shuffle=False,
                batch_size=batch_size
            )

        self.best_val_loss = np.inf
        self.best_val_acc = -np.inf
        self.valid_metrics = {} if valid_metrics is None else valid_metrics
        self.validate = False
            
    def run(self, max_steps: int=1, max_runtime=3600*24*7) -> None:
        self.start_time = datetime.datetime.now()
        self.end_time = self.start_time + datetime.timedelta(seconds=max_runtime)
        # Save initial model
        self._save_model(suffix="_initial", verbose=False)
        self._lr_nhood.clear()
        self._lr_nhood.append(self.optimizer.param_groups[0]['lr'])  # LR of the first training step
        
        while not self.terminate:
            try:
                if self.self_supervised:
                    self.km = MiniBatchKMeans(n_clusters=self.out_channels)
                    self.lr = SGDClassifier(loss='log')

                stats, misc = self._train(max_steps, max_runtime)
                self.epoch += 1

                valid_stats = self._validate()
                stats.update(valid_stats)

                self.schedulers['lr'].step(np.mean(stats['val_loss']))

                # Log to stdout and text log file
                self._log_basic(stats, misc)
                # Render visualizations and log to tensorboard
                self._log_to_tensorboard(stats, misc, None, None)

                # Save trained model state
                self._save_model(val_loss=stats['val_loss'], verbose=False) 
                # self._save_model(val_loss=stats['val_accuracy_mean'], verbose=False)  # Not verbose because it can get spammy.

                if stats['val_loss'] < self.best_val_loss:
                    self.best_val_loss = stats['val_loss']
                    self._save_model(suffix='_best', val_loss=stats['val_loss'])

                # if stats['val_accuracy_mean'] > self.best_val_acc:
                #     self.best_val_acc = stats['val_accuracy_mean']
                #     self._save_model(suffix='_best', val_loss=stats['val_accuracy_mean'])

                if self.early_stopping is not None:
                    self.early_stopping(stats["val_loss"])
                    if self.early_stopping.stop:
                        logger.info(f"Validation loss did not improve for {self.patience} epochs. Terminating...")
                        self.terminate = True

            except KeyboardInterrupt: 
                if self.ipython_shell:
                    IPython.embed(header=self._shell_info)
                else:
                    break
                if self.terminate:
                    break

            except Exception as e:
                logger.exception("Unhandled exception during training:")
                if self.ignore_errors:
                    pass
                elif self.ipython_shell:
                    print("\nEntering Command line such that Exception can be "
                          "further inspected by user.\n\n")
                    IPython.embed(header=self._shell_info)
                    if self.terminate:
                        break
                else:
                    raise e

        # Save final model
        self._save_model(suffix="_final")
        if self.tb is not None:
            self.tb.close()  # Ensure that everything is flushed

    def get_edge_attributes(self, edge_attr):
        # no edge attributes
        if self.edge_attr_ix is None:
            ea = None 
        # edge weight for GCN
        elif self.edge_attr_ix == -1: 
            ea = edge_attr[:,2] + edge_attr[:,9]
        # indexed pre-post edge attributes (e.g #incoming-#outgoing conns, incoming-outgoing area)
        else:
            ea = edge_attr[:,self.edge_attr_ix]
            assert ea.ndim == 2
            # min_value = ea.min(dim=1)[0]
            # max_value = ea.max(dim=1)[0]
            # ea -= min_value.view(-1,1)
            # diff = (max_value-min_value).view(-1, 1)
            # ea.div_(diff)
            # ea.masked_fill_(ea == float('inf'), 0)

        return ea

    def _train_step(self, batch: torch_geometric.data.Data) -> Tuple[torch.Tensor, torch.Tensor]:
        """Core training step on self.device

        Args:
            batch (torch_geometric.data.Data): batch of data

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: 
        """
        dbatch = batch.to(self.device)
        x, y = dbatch.x[:,:30], dbatch.y
        edge_index, edge_attr = dbatch.edge_index, dbatch.edge_attr
        batch_size = dbatch.batch_size
        ea = self.get_edge_attributes(edge_attr)
        
        with amp.autocast(enabled=self.mixed_precision):
            self.optimizer.zero_grad()
            # self-supervised
            if self.self_supervised:
                dloss = 0
                if self.model.encoder.__class__.__name__ == 'Linear':
                    dz = self.model.encode(x[:batch_size], edge_index)
                else:
                    dz = self.model.encode(x, edge_index, ea)[:batch_size]

                if self.embed_loss:
                    edge_index_drop, ea_drop = dropout_adj(edge_index, ea, p=0.5, force_undirected=True)
                    pos_z = self.model.encode(x, edge_index_drop, ea_drop)[:batch_size]

                    neg_ix = torch.randint(0, batch_size, size=(x.size(0),), dtype=torch.long, device=self.device)
                    neg_z = self.model.encode(x[neg_ix], edge_index, ea)[:batch_size]

                    dloss += self.model.contrastive_loss(dz, pos_z, neg_z)

                dloss += self.model.recon_loss(dz, dbatch)
                        
                z = dz.detach().cpu().numpy()
                target = y[:batch_size].detach().cpu().numpy()

                self.lr.partial_fit(z, target, classes=np.arange(self.out_channels))
                self.km.partial_fit(z)

            # semi-supervised (transductive) / supervised (inductive)
            else:  
                if self.model.__class__.__name__ == 'MLP':
                    dout = self.model(x[:batch_size])
                else:
                    ea = self.get_edge_attributes(edge_attr)
                    dout = self.model(x, edge_index, ea)[:batch_size]

                dloss = self.criterion(dout, y[:batch_size])

            dloss.backward()
            self.optimizer.step()

        if torch.isnan(dloss):
            logger.error('NaN loss detected! Aborting training.')
            raise NaNException

        return dloss

    def _train(self, max_steps, max_runtime):
        """Train for one epoch or until max_steps or max_runtime is reached"""
        self.model.train()

        # Scalar training stats that should be logged and written to tensorboard later
        stats: Dict[str, Union[float, List[float]]] = {stat: [] for stat in ['tr_loss']}
        # Other scalars to be logged
        misc: Dict[str, Union[float, List[float]]] = {misc: [] for misc in ['mean_target']}

        running_vx_size = 0  # Counts input sizes (number of pixels/voxels) of training batches

        timer = Timer()
        batch_iter = tqdm(
            enumerate(self.train_loader),
            'Training',
            total=len(self.train_loader),
            dynamic_ncols=True,
            **self.tqdm_kwargs
        )
        
        for i, batch in batch_iter:
            if self.step in self.extra_save_steps:
                self._save_model(f'_step{self.step}', verbose=False)

            dloss = self._train_step(batch)
            batch_size = batch.batch_size

            with torch.no_grad():
                loss = float(dloss)
                target = batch.y[:batch_size] if not self.self_supervised else batch.x[:batch_size]
                mean_target = float(target.to(torch.float32).mean()) if target is not None else 0.
                # misc['mean_target'].append(mean_target)
                stats['tr_loss'].append(loss)
                batch_iter.set_description(f'Training (loss {loss:.4f})')
                self._tracker.update_timeline([self._timer.t_passed, loss, mean_target])

             # Not using .get_lr()[-1] because ReduceLROnPlateau does not implement get_lr()
            misc['learning_rate'] = self.optimizer.param_groups[0]['lr']  # LR for the this iteration

            running_vx_size += batch.x[:batch_size].numel()

            self._incr_step(max_runtime, max_steps)

            if self.terminate:
                break

        stats['tr_loss_std'] = np.std(stats['tr_loss'])
        misc['tr_speed'] = len(self.train_loader) / timer.t_passed
        misc['tr_speed_vx'] = running_vx_size / timer.t_passed / 1e6  # MVx

        return stats, misc

    @torch.no_grad()
    def _validate(self) -> Tuple[Dict[str, float], Dict[str, np.ndarray]]:
        self.model.eval()  # Set dropout and batchnorm to eval mode

        val_loss = []
        lr_outs = []
        km_outs = []
        targets = []
        embeddings = []
        stats = {name: [] for name in self.valid_metrics.keys()}

        batch_iter = tqdm(
            enumerate(self.valid_loader),
            'Validating',
            total=len(self.valid_loader),
            dynamic_ncols=True,
            **self.tqdm_kwargs
        )

        for i, batch in batch_iter:
            # Everything with a "d" prefix refers to tensors on self.device (i.e. probably on GPU)
            dbatch = batch.to(self.device)
            x, y = dbatch.x[:,:30], dbatch.y
            edge_index, edge_attr = dbatch.edge_index, dbatch.edge_attr
            batch_size = dbatch.batch_size
            ea = self.get_edge_attributes(edge_attr)

            with amp.autocast(enabled=self.mixed_precision):
                # self-supervised
                if self.self_supervised:
                    if self.model.encoder.__class__.__name__ == 'Linear':
                        dz = self.model.encode(x[:batch_size], edge_index)
                    else:
                        dz = self.model.encode(x, edge_index, ea)[:batch_size]
                        
                    dloss = self.model.recon_loss(dz, dbatch)
                    z = dz.detach().cpu().numpy()
                    target = y[:batch_size].detach().cpu().numpy()

                    val_loss.append(dloss.item())

                    if self.epoch % self.validate_every_epoch == 0:
                        self.validate = True
                        # predictions
                        pred = self.lr.predict_proba(z)
                        out = self.km.predict(z)
                        lr_outs.append(pred)
                        km_outs.append(out)
                        targets.append(target)
                        embeddings.append(z)

                # semi-supervised (transductive) / supervised (inductive)
                else:
                    if self.model.__class__.__name__ == 'MLP':
                        dout = self.model(x[:batch_size])
                    else:
                        ea = self.get_edge_attributes(edge_attr)
                        dout = self.model(x, edge_index, ea)[:batch_size]

                    val_loss.append(self.criterion(dout, y[:batch_size]).item())

                    out = dout.detach().cpu().numpy()
                    target = y[:batch_size].detach().cpu().numpy()

                    if self.epoch % self.validate_every_epoch == 0:
                        self.validate = True
                        lr_outs.append(out)
                        targets.append(target)

        stats['val_loss'] = np.mean(val_loss)
        stats['val_loss_std'] = np.std(val_loss)
       
        if self.validate:
            for name, evaluator in self.valid_metrics.items():
                if evaluator.__class__.__name__ == 'SilhouetteScore':
                    mvals = [evaluator(emb, out) for emb, out in zip(embeddings, km_outs)]
                elif evaluator.__class__.__name__ in ['Accuracy', 'Precision', 'Recall', 'DSC']:  # classification evaluation
                    mvals = [evaluator(target, out) for target, out in zip(targets, lr_outs)]
                else:  # clustering evaluation
                    mvals = [evaluator(target, out) for target, out in zip(targets, km_outs)]

                if np.all(np.isnan(mvals)):
                    stats[name] = np.nan
                else:
                    stats[name] = np.nanmean(mvals)

            self.validate = False

        self.model.train()

        return stats

    def _incr_step(self, max_runtime, max_steps):
        self.step += 1

        if self.step >= max_steps:
            logger.info(f"max_steps ({max_steps}) exceeded. Terminating...")
            self.terminate = True

        if datetime.datetime.now() >= self.end_time:
            logger.info(f"max_runtime ({max_runtime} seconds) exceeded. Terminating...")
            self.terminate = True

    def _scheduler_step(self, loss):
        """Update schedules"""
        for sched in self.schedulers.values():
            # support ReduceLROnPlateau; doc. uses validation loss instead
            # http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau
            if 'metrics' in inspect.signature(sched.step).parameters:
                sched.step(metrics=loss)
            else:
                sched.step()
        # Append LR of the next iteration (after sched.step()) for local LR minima detection
        self._lr_nhood.append(self.optimizer.param_groups[0]['lr'])

    def _save_model(
        self, 
        suffix: str="",
        verbose: bool = True,
        val_loss=np.nan
    ) -> None:
        log = logger.info if verbose else logger.debug

        model = self.model
        model_trainmode = model.training

        state_dict_path = os.path.join(self.save_path, f"state_dict{suffix}.pth")
        model_path = os.path.join(self.save_path, f"model{suffix}.pt")

        info = {
            'global_step': self.step,
            'epoch': self.epoch,
            'best_val_loss': self.best_val_loss,
            'val_loss': val_loss,
            'num_neighbors': self.num_neighbors,
            'elektronn3.__version__': elektronn3.__version__,
            'env_info': collect_env.get_pretty_env_info() 
        }

        info = {k: str(v) for k, v in info.items()}

        save_dict = {
            'model_state_dict': model.state_dict(),
            'info': info,
        }

        if self.variational:
            save_dict['optimizer_state_dict'] = {
                'encoder': self.encoder_optimizer.state_dict(),
                'discriminator': self.discriminator_optimizer.state_dict(),
            }

        else:
            save_dict['optimizer_state_dict'] = self.optimizer.state_dict()

        torch.save(save_dict, state_dict_path)

        log(f'Saved state_dict as {state_dict_path}')
        pts_model_path = f"{model_path}"

        try:
            torch.save(model, model_path)
            log(f'Saved model as {model_path}')
        
        except (TypeError, PicklingError) as exc:
            raise exc

        finally:
            model.training = model_trainmode
        
        if os.path.isfile(pts_model_path):
            with zipfile.ZipFile(pts_model_path, 'a', compression=zipfile.ZIP_DEFLATED) as zfile:
                infostr = pprint.pformat(info, indent=2, width=120)
                zfile.writestr('info.txt', infostr)

    def _log_basic(self, stats, misc):
        """Log to stdout and text log file"""
        tr_loss = np.mean(stats['tr_loss'])
        val_loss = np.mean(stats['val_loss'])
        # val_acc = stats['val_accuracy_mean']
        # print(val_acc, type(val_acc))
        lr = misc['learning_rate']
        tr_speed = misc['tr_speed']
        tr_speed_vx = misc['tr_speed_vx']
        t = pretty_string_time(self._timer.t_passed)
        text = f'step={self.step:07d}, tr_loss={tr_loss:.3f}, val_loss={val_loss:.3f}, '
        # text = f'step={self.step:07d}, tr_loss={tr_loss:.3f}, val_acc={val_acc:.3f}, '
        text += f'lr={lr:.2e}, {tr_speed:.2f} it/s, {tr_speed_vx:.2f} MVx/s, {t}'
        logger.info(text)

    def _log_to_tensorboard(
        self,
        stats: Dict,
        misc: Dict,
        tr_nodes: Dict,
        val_nodes: Optional[Dict] = None,
        file_stats: Optional[Dict] = None
    ) -> None:
        if self.tb:
            try:
                self._tb_log_scalars(stats, "stats")
                self._tb_log_scalars(misc, "misc")
                if file_stats is not None:
                    self._tb_log_scalars(file_stats, "file_stats")
                self._tb_log_histograms()
            except Exception as exc:
                logger.exception('Error occured while logging to tensorboard:')
                raise exc

    def _tb_log_scalars(
            self,
            scalars: Dict[str, float],
            tag: str = 'default'
    ) -> None:
        for key, value in scalars.items():
            if isinstance(value, (list, tuple, np.ndarray)):
                for i in range(len(value)):
                    if not np.isnan(value[i]):
                        self.tb.add_scalar(f'{tag}/{key}', value[i], self.step - len(value) + i)
            elif not np.isnan(value):
                self.tb.add_scalar(f'{tag}/{key}', value, self.step)

    def _tb_log_histograms(self) -> None:
        """Log histograms of model parameters and their current gradients.

        Make sure to run this between ``backward()`` and ``zero_grad()``,
        because otherwise gradient histograms will only consist of zeros.
        """
        for name, param in self.model.named_parameters():
            self.tb.add_histogram(f'param/{name}', param, self.step)
            grad = param.grad if param.grad is not None else torch.tensor(0)
            self.tb.add_histogram(f'grad/{name}', grad, self.step)



# -*- coding: utf-8 -*-
# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Marius Killinger, Philipp Schubert, Martin Drawitsch

import logging
import time
import warnings
from typing import Dict

import matplotlib.pyplot as plt
import numpy as np
import torch

from elektronn3.training import plotting
from elektronn3 import floatX

logger = logging.getLogger('elektronn3log')


def create_preview_batch_from_knossos(knossos_preview_config: Dict[str, str]) -> torch.Tensor:
    from knossos_utils import KnossosDataset
    config = knossos_preview_config
    required_keys = ['dataset', 'size', 'offset', 'mag', 'target_mags']
    for k in required_keys:
        if k not in config.keys():
            raise ValueError(f'Required key {k} missing from knossos_preview_config.')
    logger.info(f'Loading preview region from dataset {config["dataset"]}')
    datasets = config['dataset']
    if isinstance(datasets, str):
        datasets = [datasets]
    inp_np = []
    for idx, dataset_path in enumerate(datasets):
        ds = KnossosDataset(dataset_path)
        inp_np.append(ds.load_raw(
            offset=config['offset'],
            size=config['size'],
            mag=config['mag'],
            datatype=np.float32
        ))  # (D, H, W)
    inp_np = np.stack(inp_np, axis=0) # C, D, H, W
    inp_np = inp_np[None]  # (N, C, D, H, W)
    inp_np = inp_np / config.get('scale_brightness', 1.)
    inp = torch.from_numpy(inp_np)
    return inp


class HistoryTracker:
    def __init__(self):
        self.plotting_proc = None
        self.debug_outputs = None
        self.regression_track = None
        self.debug_output_names = None

        self.timeline = AccumulationArray(n_init=int(1e5), dtype=dict(
            names=[u'time', u'loss', u'batch_char', ], formats=[u'f4', ] * 3))

        self.history = AccumulationArray(n_init=int(1e4), dtype=dict(
            names=[u'steps', u'time', u'train_loss', u'valid_loss',
                   u'loss_gain', u'train_err', u'valid_err', u'lr', u'mom',
                   u'gradnetrate'], formats=[u'i4', ] + [u'f4', ] * 9))
        self.loss = AccumulationArray(n_init=int(1e5), data=[])

    def update_timeline(self, vals):
        self.timeline.append(vals)
        self.loss.append(vals[1])

    def register_debug_output_names(self, names):
        self.debug_output_names = names

    def update_history(self, vals):
        self.history.append(vals)

    def update_debug_outputs(self, vals):
        if self.debug_outputs is None:
            self.debug_outputs = AccumulationArray(n_init=int(1e5),
                                                   right_shape=len(vals))

        self.debug_outputs.append(vals)

    def update_regression(self, pred, target):
        if self.regression_track is None:
            assert len(pred)==len(target)
            p = AccumulationArray(n_init=int(1e5), right_shape=len(pred))
            t = AccumulationArray(n_init=int(1e5), right_shape=len(pred))
            self.regression_track = [p, t]

        self.regression_track[0].append(pred)
        self.regression_track[1].append(target)

    def plot(self, save_path=None, autoscale=True, close=True, loss_smoothing_len=200):
        plotting.plot_hist(self.timeline, self.history, save_path,
                           loss_smoothing_len, autoscale)

        if self.debug_output_names and self.debug_outputs.length:
            plotting.plot_debug(self.debug_outputs, self.debug_output_names,
                                save_path)

        if self.regression_track:
            plotting.plot_regression(self.regression_track[0],
                                     self.regression_track[1], save_path)
            plotting.plot_kde(self.regression_track[0],
                              self.regression_track[1], save_path)

        if close:
            plt.close('all')


# TODO: Try to remove this thing (or document/rewrite it)
class AccumulationArray:
    def __init__(self, right_shape=(), dtype=floatX, n_init=100, data=None,
                 ema_factor=0.95):
        if isinstance(dtype, dict) and right_shape!=():
            raise ValueError("If dict is used as dtype, right shape must be"
                             "unchanged (i.e it is 1d)")

        if data is not None and len(data):
            n_init += len(data)
            right_shape = data.shape[1:]
            dtype = data.dtype

        self._n_init = n_init
        if isinstance(right_shape, int):
            self._right_shape = (right_shape,)
        else:
            self._right_shape = tuple(right_shape)
        self.dtype = dtype
        self.length = 0
        self._buffer = self._alloc(n_init)
        self._min = +np.inf
        self._max = -np.inf
        self._sum = 0
        self._ema = None
        self._ema_factor = ema_factor

        if data is not None and len(data):
            self.length = len(data)
            self._buffer[:self.length] = data
            self._min = data.min(0)
            self._max = data.max(0)
            self._sum = data.sum(0)

    def __repr__(self):
        return repr(self.data)

    def _alloc(self, n):
        if isinstance(self._right_shape, (tuple, list, np.ndarray)):
            ret = np.zeros((n,) + tuple(self._right_shape), dtype=self.dtype)
        elif isinstance(self.dtype, dict):  # rec array
            ret = np.zeros(n, dtype=self.dtype)
        else:
            raise ValueError("dtype not understood")
        return ret

    def append(self, data):
        # data = self.normalise_data(data)
        if len(self._buffer)==self.length:
            tmp = self._alloc(len(self._buffer) * 2)
            tmp[:self.length] = self._buffer
            self._buffer = tmp

        if isinstance(self.dtype, dict):
            for k, val in enumerate(data):
                self._buffer[self.length][k] = data[k]
        else:
            self._buffer[self.length] = data
            if self._ema is None:
                self._ema = self._buffer[self.length]
            else:
                f = self._ema_factor
                fc = 1 - f
                self._ema = self._ema * f + self._buffer[self.length] * fc

        self.length += 1
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            self._min = np.minimum(data, self._min)
            self._max = np.maximum(data, self._max)
            self._sum = self._sum + np.asanyarray(data)

    def add_offset(self, off):
        self.data[:] += off
        if off.ndim>np.ndim(self._sum):
            off = off[0]
        self._min += off
        self._max += off
        self._sum += off * self.length

    def clear(self):
        self.length = 0
        self._min = +np.inf
        self._max = -np.inf
        self._sum = 0

    def mean(self):
        return np.asarray(self._sum, dtype=floatX) / self.length

    def sum(self):
        return self._sum

    def max(self):
        return self._max

    def min(self):
        return self._min

    def __len__(self):
        return self.length

    @property
    def data(self):
        return self._buffer[:self.length]

    @property
    def ema(self):
        return self._ema

    def __getitem__(self, slc):
        return self._buffer[:self.length][slc]


class Timer:
    def __init__(self):
        self.origin = time.time()
        self.t0 = self.origin

    @property
    def t_passed(self):
        return time.time() - self.origin


def pretty_string_time(t):
    """Custom printing of elapsed time"""
    if t > 4000:
        s = 't=%.1fh' % (t / 3600)
    elif t > 300:
        s = 't=%.0fm' % (t / 60)
    else:
        s = 't=%.0fs' % (t)
    return s



# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Martin Drawitsch

"""Neural network layers"""

from typing import Optional, Tuple

import torch
from torch import nn


class GatherExcite(nn.Module):
    """Gather-Excite module (https://arxiv.org/abs/1810.12348),

    a generalization of the Squeeze-and-Excitation module
    (https://arxiv.org/abs/1709.01507).

    Args:
        channels: Number of input channels (= number of output channels)
        extent: extent factor that determines how much the gather operator
            output is smaller than its input. The special value ``extent=0``
            activates global gathering (so the gathered information has no
            spatial extent).
        param_gather: If ``True``, the gather operator is parametrized
            according to https://arxiv.org/abs/1810.12348.
        param_excite: If ``True``, the excitation operator is parametrized
            according to https://arxiv.org/abs/1810.12348 (also equivalent to
            the original excitation operator proposed in
            https://arxiv.org/abs/1709.01507).
        reduction:  Channel reduction rate of the parametrized excitation
            operator.
        spatial_shape: Spatial shape of the module input. This needs to be
            specified if ``param_gather=0 and extent=0`` (parametrized global
            gathering).
    """
    def __init__(
            self,
            channels: int,
            extent: int = 0,
            param_gather: bool = False,
            param_excite: bool = True,
            reduction: int = 16,
            spatial_shape: Optional[Tuple[int, ...]] = None
    ):
        super().__init__()
        if extent == 1:
            raise NotImplementedError('extent == 1 doesn\'t make sense.')
        if param_gather:
            if extent == 0:  # Global parametrized gather operator
                if spatial_shape is None:
                    raise ValueError(
                        'With param_gather=True, extent=0, you will need to specify spatial_shape.')
                self.gather = nn.Sequential(
                    nn.Conv3d(channels, channels, spatial_shape),
                    nn.BatchNorm3d(channels),
                    nn.ReLU()
                )
            else:
                # This will make the model much larger with growing extent!
                # TODO: This is ugly and I'm not sure if it should even be supported
                assert extent in [2, 4, 8, 16]
                num_convs = int(torch.log2(torch.tensor(extent, dtype=torch.float32)))
                self.gather = nn.ModuleList([
                    nn.Sequential(
                        nn.Conv3d(channels, channels, 3, stride=2, padding=1),
                        nn.BatchNorm3d(channels),
                        nn.ReLU()
                    ) for _ in range(num_convs)
                ])
        else:
            if extent == 0:
                self.gather = nn.AdaptiveAvgPool3d(1)  # Global average pooling
            else:
                self.gather = nn.AvgPool3d(extent)
        if param_excite:
            self.excite = nn.Sequential(
                nn.Conv3d(channels, channels // reduction, 1),
                nn.ReLU(),
                nn.Conv3d(channels // reduction, channels, 1)
            )
        else:
            self.excite = nn.Identity()

        if extent == 0:
            self.interpolate = nn.Identity()  # Use broadcasting instead of interpolation
        else:
            self.interpolate = torch.nn.functional.interpolate

    def forward(self, x):
        y = self.gather(x)
        y = self.excite(y)
        y = torch.sigmoid(self.interpolate(y, x.shape[2:]))
        return x * y



# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany

import torch
import torch.nn as nn
import glob
from collections import OrderedDict
import numpy as np
import time
from typing import Union
from elektronn3.training.train_utils import pretty_string_time


class InferenceModel(object):
    """Class to perform inference using a trained elektronn3 model or nn.Module object.

    Args:
        src: Path to training folder of e3 model or already loaded/initialized nn.Module defining the model.
        disable_cuda: use cpu only
        multi_gpu: enable multi-gpu support of pytorch
    Examples:
        >>> cnn = nn.Sequential(
        ... nn.Conv2d(5, 32, 3, padding=1), nn.ReLU(),
        ... nn.Conv2d(32, 2, 1)).to('cpu')
        >>> inp = np.random.randn(2, 5, 10, 10)
        >>> model = InferenceModel(cnn)
        >>> out = model.predict_proba(inp)
        >>> assert np.all(np.array(out.shape) == np.array([2, 2, 10, 10]))
    """
    def __init__(self, src: Union[str, nn.Module], disable_cuda: bool = False,
                 multi_gpu: bool = True, normalize_func=None):
        self.normalize_func = normalize_func
        if not disable_cuda and torch.cuda.is_available():
            device = torch.device('cuda')
        else:
            device = torch.device('cpu')
        self.device = device
        if type(src) is str:
            print('Initializing model from {}.'.format(src))
            self.model = load_model(src)
            self.model_p = src
        else:
            self.model = src
            self.model_p = None
        self.model.eval()
        if multi_gpu:
            self.model = nn.DataParallel(self.model)
        self.model.to(self.device)

    def predict_proba(self, inp: np.ndarray, bs: int = 10,
                        verbose: bool = False):
        """

        Args:
            inp: Input data, e.g. of shape [N, C, H, W]
            bs: batch size
            verbose: report inference speed

        Returns:

        """
        if verbose:
            start = time.time()
        if self.normalize_func is not None:
            inp = self.normalize_func(inp)
        with torch.no_grad():
            # get output shape shape
            if type(inp) is tuple:
                out = self.model(*(torch.Tensor(ii[:2]).to(torch.float32).to(self.device) for ii in inp))
                n_samples = len(inp[0])
            else:
                out = self.model(torch.Tensor(inp[:2]).to(torch.float32).to(self.device))
                n_samples = len(inp)
            # change sample number according to input
            if type(out) is tuple:
                out = tuple(np.zeros([n_samples] + list(out[ii].shape)[1:],
                               dtype=np.float32) for ii in range(len(out)))
            else:
                out = np.zeros([n_samples] + list(out.shape)[1:], dtype=np.float32)
            for ii in range(0, int(np.ceil(n_samples / bs))):
                low = bs * ii
                high = bs * (ii + 1)
                if type(inp) is tuple:
                    inp_stride = tuple(torch.Tensor(ii[low:high]).to(torch.float32).to(self.device) for ii in inp)
                    res = self.model(*inp_stride)
                else:
                    inp_stride = torch.Tensor(inp[low:high]).to(torch.float32).to(self.device)
                    res = self.model(inp_stride)
                if type(res) is tuple:
                    for ii in range(len(res)):
                        out[ii][low:high] = res[ii].detach().cpu()
                else:
                    out[low:high] = res.detach().cpu()
                if type(inp_stride) is tuple:
                    for el in inp_stride:
                        el.detach_()
                else:
                    inp_stride.detach_()
                del inp_stride
                del res
                torch.cuda.empty_cache()
            assert high >= n_samples, "Prediction less samples then given" \
                                     " in input."
        if verbose:
            dtime = time.time() - start
            if type(inp) is tuple:
                inp_el = np.sum([float(np.prod(inp[kk].shape)) for kk in range(len(inp))])
            else:
                inp_el = float(np.prod(inp.shape))
            speed = inp_el / dtime / 1e6
            dtime = pretty_string_time(dtime)
            print(f'Inference speed: {speed:.2f} MB or MPix /s, time: {dtime}.')
        return out


def load_model(src: str) -> nn.Module:
    """
    Load trained elektronn3 model.

    Args:
        src: Source path to model directory. Directory must contain training
        script and model-checkpoint.pth.

    Returns:
        Trained model
    """
    train_script = glob.glob(f'{src}/*.py')
    assert len(train_script) == 1, "Multiple/None trainer file(s). " \
                                   "Ill-defined trainer script."
    exec(open(train_script[0]).read(), globals())
    assert "get_model" in globals(), "'get_model' not defiend in trainer script."
    model = get_model()
    state_dict_p = glob.glob(f'{src}/*.pth')
    if len(state_dict_p) > 1:
        final_p = ["final" in sp for sp in state_dict_p]
        if np.sum(final_p) == 1:
            state_dict_p = [state_dict_p[np.argmax(final_p)]]
    assert len(state_dict_p) == 1, "Multiple/None state dict file(s). " \
                                   "Ill-defined state dict file."
    state_dict = torch.load(state_dict_p[0])
    try:
        model.load_state_dict(state_dict)
    # if model was saved as nn.DataParallel then remove 'module.'
    #  prefix in every key
    except RuntimeError:
        new_state_dict = OrderedDict()
        for k, v in state_dict.items():
            new_state_dict[k.replace('module.', '')] = v
        try:  # HACK to ensure backwards compat. with example cube models TODO: remove ASAP
            model.load_state_dict(new_state_dict)
        except RuntimeError:
            newnew_state_dict = OrderedDict()
            for k, v in new_state_dict.items():
                newnew_state_dict[k.replace('pretrained_net.', 'base_net.')] = v
            model.load_state_dict(newnew_state_dict)
    return model.eval()



"""
Trainer code for 2D and 3D Noise2Void (https://arxiv.org/abs/1811.10980)
Adapted from https://github.com/juglab/pn2v/blob/master/pn2v/training.py,
ported from NumPy to PyTorch and generalized to support 3D.
"""

from typing import Callable

import torch
from torch import nn
from torch.cuda import amp
import numpy as np
import itertools

from scipy.ndimage.filters import gaussian_filter
from tqdm import tqdm

from elektronn3.training.trainer import Trainer, NaNException
from elektronn3.modules.loss import MaskedMSELoss

import logging
logger = logging.getLogger('elektronn3log')


@torch.no_grad()
def get_stratified_coords(ratio, shape):
    """
    Produce a list of approx. ``num_pix`` random coordinates, sampled from
    ``shape`` using startified sampling. Supports n-dimensional shapes.
    """
    # total_num = torch.prod(shape).to(torch.float32)
    # sample_num = total_num * ratio
    ratio = torch.as_tensor(ratio)
    ndim = len(shape)
    shape = torch.as_tensor(shape, dtype=torch.int32)
    box_size = int(torch.round(torch.sqrt(1. / ratio)))
    coords = []
    box_counts = torch.ceil(shape.float() / box_size).int()
    for steps in itertools.product(*[range(bc) for bc in box_counts]):
        steps = torch.as_tensor(steps, dtype=torch.int32)
        co = torch.randint(0, box_size, (ndim,)) + box_size * steps
        if torch.all(co < shape):
            coords.append(co)
    if not coords:
        raise ValueError(f'ratio {ratio:.1e} is too close to zero. Choose a higher value.')
    coords = torch.stack(coords)
    return coords


# TODO: Is the hardcoded small ROI size sufficient?
@torch.no_grad()
def prepare_sample(img, ratio=1e-3, channels=None):
    """Prepare binary mask and target image for Noise2Void from a given image"""
    ndim = img.ndim - 2  # Subtract (N, C) dims
    if channels is None:
        channels = range(img.shape[1])
    inp = img.clone()
    target = img
    mask = torch.zeros_like(img)
    for n, c in itertools.product(range(img.shape[0]), channels):
        hotcoords = get_stratified_coords(ratio, img[n, c].shape)
        maxsh = np.array(img[n, c].shape) - 1
        for hc in hotcoords:
            roimin = np.clip(hc - 2, 0, None)
            roimax = np.clip(hc + 3, None, maxsh)
            roi = img[n, c, roimin[0]:roimax[0], roimin[1]:roimax[1]]
            if ndim == 3:
                roi = roi[..., roimin[2]:roimax[2]]  # slice 3rd dim if input is 3D
            rc = np.full((ndim,), 2)
            while np.all(rc == 2):
                rc = np.random.randint(0, roi.shape, (ndim,))
            repl = roi[tuple(rc)]  # Select point at rc in current ROI for replacement
            inp[(n, c, *hc)] = repl
            mask[(n, c, *hc)] = 1.0

    return inp, target, mask


class Noise2VoidTrainer(Trainer):
    """Trainer subclass with custom training and validation code for Noise2Void training.

    Noise2Void is applied by default, but it can also be replaced or accompanied by additive
    gaussian noise and gaussian blurring (see args below).

    Args:
        model: PyTorch model (``nn.Module``) that shall be trained.
        criterion: Training criterion. If ``n2v_ratio > 0``, it should expect 3 arguments,
            the third being the Noise2Void mask. Per default, a masked MSE loss is used.
        *args: *Other positional args. See signature of :py:class:`elektronn3.training.Trainer`*
        n2v_ratio: Ratio of pixels to be manipulated and masked in each image according to the
            Noise2Void algorithm. If it is set to a value <= 0, Noise2Void is disabled.
        agn_max_std: Maximum std (sigma parameter) for additive gaussian noise that is
            optionally applied to the input image. Standard deviations are sampled from a uniform
            distribution that ranges between 0 and ``agn_max_std``.
            If it is set to a value <= 0, additive gaussian noise is disabled.
        gblur_sigma: Sigma parameter for gaussian blurring that is optionally applied to the
            input image. If it is set to a value <= 0, gaussian blurring is disabled.
        **kwargs: Other keyword args. See signature of :py:class:`elektronn3.training.Trainer`
    """
    def __init__(
            self,
            model: torch.nn.Module,
            criterion: torch.nn.Module = MaskedMSELoss(),
            *args,
            n2v_ratio: float = 1e-3,
            agn_max_std: float = 0,
            gblur_sigma: float = 0,
            **kwargs
    ):
        super().__init__(model, criterion, *args, **kwargs)
        self.n2v_ratio = n2v_ratio
        self.agn_max_std = agn_max_std
        self.gblur_sigma = gblur_sigma

    def _train_step(self, batch):
        # Everything with a "d" prefix refers to tensors on self.device (i.e. probably on GPU)
        dimg = batch['inp'].to(self.device, non_blocking=True)

        if self.n2v_ratio > 0:
            dinp, dtarget, dmask = prepare_sample(dimg, ratio=self.n2v_ratio)
        else:
            dinp = dimg.clone()
            dtarget = dimg
            dmask = None

        # Apply additive gaussian noise
        if self.agn_max_std > 0:
            agn_std = np.random.rand() * self.agn_max_std  # stds from range [0, agn_max_std]
            dinp.add_(torch.randn_like(dinp).mul_(agn_std))

        # Apply gaussian blurring
        if self.gblur_sigma > 0:
            dinp = dinp.cpu().numpy()
            for n, c in itertools.product(range(dinp.shape[0]), range(dinp.shape[1])):
                dinp[n, c] = gaussian_filter(dinp[n, c], sigma=self.gblur_sigma)
            dinp = torch.as_tensor(dinp).to(self.device).float()

        # forward pass
        with amp.autocast(enabled=self.mixed_precision):
            dout = self.model(dinp)
            if dmask is None:
                dloss = self.criterion(dout, dtarget)
            else:
                dloss = self.criterion(dout, dtarget, dmask)
        if torch.isnan(dloss):
            logger.error('NaN loss detected! Aborting training.')
            raise NaNException
        # update step
        self.scaler.scale(dloss).backward()
        self.scaler.step(self.optimizer)
        self.scaler.update()
        self.optimizer.zero_grad(set_to_none=True)
        return dloss, dout

    @torch.no_grad()
    def _validate(self):
        self.model.eval()  # Set dropout and batchnorm to eval mode

        val_loss = []
        outs = []
        targets = []
        stats = {name: [] for name in self.valid_metrics.keys()}
        batch_iter = tqdm(
            enumerate(self.valid_loader),
            'Validating',
            total=len(self.valid_loader),
            dynamic_ncols=True,
            **self.tqdm_kwargs
        )
        for i, batch in batch_iter:
            dimg = batch['inp'].to(self.device, non_blocking=True)

            if self.n2v_ratio > 0:
                dinp, dtarget, dmask = prepare_sample(dimg, ratio=self.n2v_ratio)
            else:
                dinp = dimg.clone()
                dtarget = dimg
                dmask = None

            # Apply additive gaussian noise
            if self.agn_max_std > 0:
                agn_std = np.random.rand() * self.agn_max_std  # stds from range [0, agn_max_std]
                dinp.add_(torch.randn_like(dinp).mul_(agn_std))

            # Apply gaussian blurring
            if self.gblur_sigma > 0:
                dinp = dinp.cpu().numpy()
                for n, c in itertools.product(range(dinp.shape[0]), range(dinp.shape[1])):
                    dinp[n, c] = gaussian_filter(dinp[n, c], sigma=self.gblur_sigma)
                dinp = torch.as_tensor(dinp).to(self.device).float()

            # forward pass
            with amp.autocast(enabled=self.mixed_precision):
                dout = self.model(dinp)
                if dmask is None:
                    dloss = self.criterion(dout, dtarget)
                else:
                    dloss = self.criterion(dout, dtarget, dmask)
            val_loss.append(dloss.item())
            out = dout.detach().cpu()
            outs.append(out)
            targets.append(dtarget)
        images = {
            'inp': dinp.cpu().numpy(),
            'out': dout.cpu().numpy(),
            'target': None if dtarget is None else dtarget.cpu().numpy(),
            'fname': batch.get('fname'),
        }
        self._put_current_attention_maps_into(images)

        stats['val_loss'] = np.mean(val_loss)
        stats['val_loss_std'] = np.std(val_loss)

        for name, evaluator in self.valid_metrics.items():
            mvals = [evaluator(target, out) for target, out in zip(targets, outs)]
            if np.all(np.isnan(mvals)):
                stats[name] = np.nan
            else:
                stats[name] = np.nanmean(mvals)

        self.model.train()  # Reset model to training mode

        return stats, images


if __name__ == '__main__':
    # Demo of Noise2Void training sample generation
    import matplotlib.pyplot as plt
    import scipy.misc

    # co = get_stratified_coords(16, (8, 8, 3))
    # print(co)
    im = scipy.misc.ascent()[::2, ::2]
    imt = torch.as_tensor(im)[None, None]
    inp, target, mask = prepare_sample(imt, 1e-3)
    fig, axes = plt.subplots(ncols=3, constrained_layout=True, figsize=(20, 12))
    axes[0].imshow(im, cmap='gray')
    axes[0].set_title('Original image')
    axes[1].imshow(mask[0,0])
    axes[1].set_title('Mask')
    axes[2].imshow(inp[0,0], cmap='gray')
    axes[2].set_title('Manipulated image for Noise2Void training')
    plt.show()



import os
import versioneer

from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    install_requires = []
else:
    install_requires = [
        'torch',
        'numpy',
        'scipy',
        'h5py',
        'matplotlib',
        'numba',
        'ipython',
        'imageio',
        'colorlog',
        'tqdm',
        'scikit-learn',
        'scikit-image',
        'tensorboardX',
        'torchvision'
    ]

setup(
    name='elektronn3',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Utilities for 3D CNNs in PyTorch',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/ELEKTRONN/elektronn3',
    author='ELEKTRONN team',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    packages=find_packages(exclude=['scripts']),
    install_requires=install_requires,
)



import os
import random
from typing import Callable, Dict, Optional, Sequence

import knossos_utils
import numpy as np
import torch
import torch.utils.data

from elektronn3.data import transforms


class KnossosRawData(torch.utils.data.Dataset):
    """Delivers raw patches that are randomly sampled from a KNOSSOS dataset.
    Supports 2D and 3D (choice is determined from the length of the
    ``patch_shape`` param).

    Args:
        conf_path: Path to KNOSSOS .conf file
        patch_shape: Shape (zyx) of patches that should be sampled from the
            dataset.
        transform: Transformation to be applied to the loaded data.
            See :py:mod:`elektronn3.data.transforms`.
        bounds: Tuple of boundary coordinates (xyz) that constrain the region
            of where patches should be sampled from within the KNOSSOS dataset.
            E.g. ``bounds=((256, 256, 128), (512, 512, 512))`` means that only
            the region between the low corner (x=256, y=256, z=128) and the
            high corner (x=512, y=512, z=512) of the dataset is considered. If
            ``None``, the whole dataset is used.
        mag: KNOSSOS magnification number
        mode: One of ``in_memory``, ``caching`` or ``disk``. If ``in_memory`` (default), the dataset (or the subregion
            that is constrained by ``bounds``) is pre-loaded into memory on initialization. If ``caching``, cache data
            from the disk and reuse it. If ``disk``, load data from disk on demand.
        epoch_size: Determines the length (``__len__``) of the ``Dataset``
            iterator. ``epoch_size`` can be set to an arbitrary value and
            doesn't have any effect on the content of produced training
            samples.
        disable_memory_check: If ``False`` (default), the amount of required
            memory is compared to the amount of free memory. If a planned
            allocation that exceeds 90% of free memory is detected, an error
            is raised. If ``True``, this check is disabled.
        verbose: If ``True``, be verbose about disk I/O.
        cache_size: If mode=``cache``: How many samples to hold in cache.
        cache_reuses: If mode=``cache``: How often to reuse a sample in cache before loading a new one from disk.
    """

    def __init__(
            self,
            conf_path: str,
            patch_shape: Sequence[int],  # [z]yx
            transform: Callable = transforms.Identity(),
            bounds: Optional[Sequence[Sequence[int]]] = None,  # xyz
            mag: int = 1,
            mode: str = 'in_memory',
            epoch_size: int = 100,
            disable_memory_check: bool = False,
            verbose: bool = False,
            cache_size: int = 50,
            cache_reuses: int = 10
    ):
        self.conf_path = conf_path
        self.patch_shape = np.array(patch_shape)
        self.transform = transform
        self.mag = mag
        self.epoch_size = epoch_size
        self.disable_memory_check = disable_memory_check
        self.verbose = verbose
        self.cache_size = cache_size
        self.cache_reusages = cache_reuses
        if mode not in ['in_memory', 'caching', 'disk']:
            raise ValueError(f'mode has to be one of ``in_memory``, ``caching`` or ``disk``, but is {mode}')
        self.mode = mode

        self.kd = knossos_utils.KnossosDataset(self.conf_path, show_progress=self.verbose)
        self.dim = len(self.patch_shape)
        patch_shape_xyz = self.patch_shape[::-1]  # zyx -> xyz
        if self.dim == 2:
            patch_shape_xyz = np.array([*patch_shape_xyz, 1])  # z=1 for 2D
        self.patch_shape_xyz = patch_shape_xyz
        if bounds is None:
            bounds = [[0, 0, 0], self.kd.boundary]
        self.bounds = np.array(bounds)
        self.shape = self.bounds[1] - self.bounds[0]
        self.raw = None  # Will be filled with raw data if in_memory is True

        if self.mode == 'in_memory':
            self._load_into_memory()
        elif self.mode == 'caching':
            self._fill_cache()

    def __getitem__(self, index: int) -> Dict[str, torch.Tensor]:
        if self.mode == 'in_memory':
            inp, offset = self._load_from_memory()
        elif self.mode == 'caching':
            inp, offset = self._get_from_cache()
        else:
            inp, offset = self._load_from_disk()
        if self.dim == 2:
            inp = inp[0]  # squeeze z=1 dim -> yx
        inp = inp.astype(np.float32)[None]  # Prepend C dim -> (C, [D,] H, W)
        if self.transform:
            inp, _ = self.transform(inp, None)
        sample = {
            'inp': torch.as_tensor(inp),
            'offset': offset
        }
        return sample

    def _load_from_disk(self) -> (np.ndarray, np.ndarray):
        min_offset = self.bounds[0]  # xyz
        max_offset = self.bounds[1] - self.patch_shape_xyz  # xyz
        offset = np.random.randint(min_offset, max_offset + 1)  # xyz
        inp = self.kd.load_raw(
            offset=offset, size=self.patch_shape_xyz, mag=self.mag
        )  # zyx (D, H, W)
        return inp, offset

    def _load_from_memory(self) -> (np.ndarray, np.ndarray):
        min_offset = (0, 0, 0)  # 0 because self.raw already accounts for min offset
        max_offset = self.shape - self.patch_shape_xyz
        offset = np.random.randint(min_offset, max_offset + 1)  # xyz
        inp = self.raw[  # self.raw has zyx dim order
              offset[2]:offset[2] + self.patch_shape_xyz[2],
              offset[1]:offset[1] + self.patch_shape_xyz[1],
              offset[0]:offset[0] + self.patch_shape_xyz[0],
        ]  # zyx (D, H, W)
        return inp, offset

    def __len__(self) -> int:
        return self.epoch_size

    def _load_into_memory(self) -> None:
        if not self.disable_memory_check:
            self.memory_check()
        if self.verbose:
            print('Loading dataset into memory...')
        self.raw = self.kd.load_raw(
            offset=self.bounds[0], size=self.shape, mag=self.mag
        )  # zyx (D, H, W)

    def memory_check(self) -> None:
        # Memory cost in GiB: Number of gibivoxels * 4 because each uint8 voxel needs 1 byte.
        required_mem_gib = np.prod(self.bounds[1] - self.bounds[0]) / 1024 ** 3 * 1
        # Total RAM on the system in GiB
        free_mem_gib = float(os.popen('free -tm').readlines()[-1].split()[3]) / 1024
        mem_frac = required_mem_gib / free_mem_gib
        if mem_frac > 0.9:
            raise RuntimeError(
                f'Using in_memory and bounds of {self.bounds.tolist()} would require at least '
                f'{required_mem_gib:.2f} GiB of memory for each dataset instance.\n'
                'Please specifiy smaller bounds, or if you are sure about what you are '
                'doing, you can disable this check with the disable_memory_check flag.'
            )

    def _fill_cache(self) -> None:
        self.cache = [self._load_from_disk() for _ in range(self.cache_size)]
        self.remaining_cache_reusages = [self.cache_reusages] * self.cache_size

    def _get_from_cache(self) -> (np.ndarray, np.ndarray):
        idx = random.randrange(self.cache_size)
        inp = self.cache[idx][0]
        offset = self.cache[idx][1]
        self.remaining_cache_reusages[idx] -= 1
        if self.remaining_cache_reusages[idx] < 1:
            self.cache[idx] = self._load_from_disk()
            self.remaining_cache_reusages[idx] = self.cache_reusages
        return inp, offset
